#ifndef __DEBUG_H
#define __DEBUG_H

#include <assert.h>
#include "boolean.h"

/* CODE-VERIFYING FEATURES:
   The following macros can be used to insert code which ensures
   that some specific requirements are met at some point of the
   execution. Such a code is only compiled when the flag NDEBUG 
   is undefined and is ignored otherwise.
*/

#ifndef NDEBUG /* Activate assertion features */

#define CHECK(cond) assert(cond) 
/* Check the condition 'cond' and abort execution if 'cond' is false */

#define CCODE(code) code
/* Introduce assertion-code 'code' */

#define CFAIL() CHECK(FALSE)
/* Always fail and abort execution */

#define GUARD(cond,action) (CHECK(cond),(action))
/* Check condition 'cond' before executing 'action'.
   The execution is aborted if 'cond' is false. */

#else /* Inhibit assertion features */

#define CHECK(cond) /* Do nothing */
#define CCODE(code) /* Do nothing */
#define CFAIL() /* Do nothing */
#define GUARD(cond,action) (action) /* Do 'action' without checking 'cond' */

#endif

#define REQUIRE(cond) CHECK(cond)
#define ENSURE(cond)  CHECK(cond)
#define GETSAFE(cond,val) GUARD((cond),(val))
#define SETSAFE(cond,var,val) GUARD((cond),(var) = (val))
		  
#define VALUE(v) GUARD(TRUE,(v))
/* Makes 'v' a read-only rvalue only, thus preventing it to be used as 
   an lvalue (left side of an assignment) */


/* DEBUGGING FEATURES:
   The following macros can be used to insert code which facilitates
   the debugging task. This code is only compiled when the flag NDEBUG
   (no debug) is undefined. Otherwise, the code is compiled but is 
   executed only if the level associated to it via the DEBUG macro is 
   lower than or equal to the value specified in DDEPTH (debug depth).
*/

#ifndef NDEBUG /* Activate debugging features */

#ifndef DDEPTH
#define DDEPTH 0 /* Set debug depth to 0 if it is not already defined */
#endif
#define IFDBG(depth,debcode,code) do if ((depth)<=DDEPTH) (debcode); while(0)
#define DEBUG(depth,debcode) IFDBG((depth),(debcode),TRUE)

#else /* Inhibit debugging features */

#define IFDBG(depth,debcode,code) (code) 
#define DEBUG(depth,code) 

#endif

#endif /* __DEBUG_H */
