/* This module defines a fifo data-structure of generic pointers (void *) */
 
#ifndef __FIFO_H
#define __FIFO_H

#include "platform.h"
#include "index.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* Definition of the FIFO data-structure */

typedef	struct fifo_s fifo_t, *FIFO;
struct fifo_s
{                               
  ITEM *fi_item; /* Array of generic pointers to the elements */
  CSIZE fi_capa; /* Capacity of the fifo */
  CSIZE fi_size; /* Number of elements in the fifo */
  CINDX fi_tail; /* Index of the last element of the fifo */
  CINDX fi_head; /* Index of the first element of the fifo */
};  

/* Clean access to fifo-structure fields */

#define FIFO_ITEM(fifo) ((fifo)->fi_item)
#define FIFO_CAPA(fifo) ((fifo)->fi_capa)
#define FIFO_SIZE(fifo) ((fifo)->fi_size)
#define FIFO_TAIL(fifo) ((fifo)->fi_tail)
#define FIFO_HEAD(fifo) ((fifo)->fi_head)

/*************
 * INTERFACE *
 *************/

/* ASSERTIONS */

#define IS_FIFO(fifo) ((fifo) != NULL)

/* CONSTRUCTORS - DESTRUCTORS */

public FIFO NewFifo(CSIZE capacity);
/* Create a new fifo structure with capacity 'capacity'
   Return:
     - A pointer to a new fifo if the function succeeds
     - NULL if the function fails (not enough memory)
   Parameters:
     'capacity' is the maximum number of elements the fifo 
     will be able to hold
   Remarks:
     If 'capacity' is less than one, the function creates a 
     fifo whose capacity is exactly one.
*/

public void DelFifo(FIFO fifo);
/* Delete the fifo 'fifo'
   Parameter: 'fifo' is the fifo which should be deleted
   Remark: Nothing is done if 'fifo' is NULL.
*/


/* ACCESSORS */

#define FifoCapa(fifo) GETSAFE(IS_FIFO(fifo),FIFO_CAPA(fifo))
/* Return: the maximum number of elements 'fifo' can hold */

#define FifoSize(fifo) GETSAFE(IS_FIFO(fifo),FIFO_SIZE(fifo))
/* Return: the actual number of elements in 'fifo' */

#define FifoIsEmpty(fifo) (FifoSize(fifo) == CSIZE_C(0))
/* Return: TRUE iff 'fifo' has no elements */

#define FifoIsFull(fifo) (FifoSize(fifo) == FifoCapa(fifo))
/* Return: TRUE iff 'fifo' has as many elements as its capacity */

#define FifoLastItem(fifo) GETSAFE(! FifoIsEmpty(fifo),\
                                   FIFO_ITEM(fifo)[FIFO_TAIL(fifo)])
/* Return: the last element of 'fifo'
   Require: ! FifoIsEmpty(fifo)
*/ 

#define FifoFirstItem(fifo) GETSAFE(! FifoIsEmpty(fifo),\
                                    FIFO_ITEM(fifo)[FIFO_HEAD(fifo)])
/* Return: the first element of 'fifo'
   Require: ! FifoIsEmpty(fifo) 
*/


/* MODIFIERS */

public void FifoPutLast(FIFO fifo,ITEM item);
/* Put the element 'item' at the end of the fifo 'fifo', 
   thus making it the last element in the fifo
   Parameters:
     - 'fifo' is the fifo in which 'item' should be put
     - 'item' is the element to add 
   Require: ! FifoIsFull(fifo)
*/
#define FifoPush(fifo,elem) FifoPutLast((fifo),(elem))

public ITEM FifoRemFirst(FIFO fifo);
/* Remove the first element of 'fifo'
   Parameter: 'fifo' is the fifo to get from
   Return: the first element of 'fifo'
   Require: ! FifoIsEmpty(fifo)
*/
#define FifoPop(fifo) FifoRemFirst(fifo)

public void FifoReset(FIFO lifo);
/* Remove all elements of 'fifo'
   Parameter: 'fifo' is the fifo to reset
   Ensure: FifoIsEmpty(lifo)
*/

#endif





