/* This module defines a lifo data-structure of generic pointers (void *) */
 
#ifndef __LIFO_H
#define __LIFO_H

#include "platform.h"
#include "index.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* Definition of the LIFO data-structure */

typedef	struct lifo_s lifo_t, *LIFO;
struct lifo_s
{                               
  ITEM *li_item; /* Array of generic pointers to the elements */
  CSIZE li_capa; /* Capacity of the lifo */
  CSIZE li_size; /* Number of elements in the lifo */
  CINDX li_tail; /* Index of the last element of the lifo */
};  

/* Clean access to lifo-structure fields */

#define LIFO_ITEM(lifo) ((lifo)->li_item)
#define LIFO_CAPA(lifo) ((lifo)->li_capa)
#define LIFO_SIZE(lifo) ((lifo)->li_size)
#define LIFO_TAIL(lifo) ((lifo)->li_tail)

/*************
 * INTERFACE *
 *************/

/* ASSERTIONS */

#define IS_LIFO(lifo) ((lifo) != NULL)

/* CONSTRUCTORS - DESTRUCTORS */

public LIFO NewLifo(CSIZE capacity);
/* Create a new lifo structure with capacity 'capacity'
   Return:
     - A pointer to a new lifo if the function succeeds
     - NULL if the function fails (not enough memory)
   Parameters:
     'capacity' is the maximum number of elements the lifo 
     will be able to hold
   Remarks:
     If 'capacity' is less than one, the function creates a 
     lifo whose capacity is exactly one.
*/

public void DelLifo(LIFO lifo);
/* Delete the lifo 'lifo'
   Parameter: 'lifo' is the lifo which should be deleted
   Remark: Nothing is done if 'lifo' is NULL.
*/

/* ACCESSORS */

#define LifoCapa(lifo) GETSAFE(IS_LIFO(lifo),LIFO_CAPA(lifo))
/* Return: the maximum number of elements 'lifo' can hold */

#define LifoSize(lifo) GETSAFE(IS_LIFO(lifo),LIFO_SIZE(lifo))
/* Return: the actual number of elements in 'lifo' */

#define LifoIsEmpty(lifo) (LifoSize(lifo) == CSIZE_C(0))
/* Return: TRUE iff 'lifo' has no elements */

#define LifoIsFull(lifo) (LifoSize(lifo) == LifoCapa(lifo))
/* Return: TRUE iff 'lifo' has as many elements as its capacity */

#define LifoItem(lifo,idx) GETSAFE(IS_LIFO(lifo) && \
                                   RANGE(CINDX_C(0),(idx),LIFO_TAIL(lifo)),\
                                   LIFO_ITEM(lifo)[idx])

#define LifoLastItem(lifo) LifoItem(IS_LIFO(lifo),LIFO_TAIL(lifo))
/* Return: the last element of 'lifo'
   Require: ! LifoIsEmpty(lifo)
*/ 
#define LifoTop(lifo) LifoLastItem(lifo)


/* MODIFIERS */

public void LifoPutLast(LIFO lifo,ITEM item);
/* Put the element 'item' at the end of the lifo 'lifo', 
   thus making it the last element in the lifo
   Parameters:
     - 'lifo' is the lifo in which 'item' should be put
     - 'item' is the element to add 
   Require: ! LifoIsFull(lifo)
*/
#define LifoPush(lifo,item) LifoPutLast((lifo),(item))

public ITEM LifoRemLast(LIFO lifo);
/* Remove the last element of 'lifo'
   Parameter: 'lifo' is the lifo to get from
   Return: the last element of 'lifo'
   Require: ! LifoIsEmpty(lifo)
*/
#define LifoPop(lifo) LifoRemLast(lifo)

public void LifoReset(LIFO lifo);
/* Remove all elements of 'lifo'
   Parameter: 'lifo' is the lifo to reset
   Ensure: LifoIsEmpty(lifo)
*/

#endif





