#include "list.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* HELPERS */

#define NEW_LNODE(newnode,item) \
{ \
  newnode = (LNODE) malloc(sizeof(struct lnode_s)); \
  if (! newnode) return FALSE; \
  LNODE_ITEM(newnode) = item; \
}

#define DEL_LNODE(node) \
{ \
  if (node) free(node) ; \
}

#define LINK_LNODE(leftnode,rightnode,node) \
{ \
  LNODE_SUCC(node) = rightnode; \
  LNODE_PRED(node) = leftnode; \
  LNODE_SUCC(leftnode) = node; \
  LNODE_PRED(rightnode) = node; \
}

#define UNLINK_LNODE(node) \
{\
  LNODE_SUCC(LNODE_PRED(node)) = LNODE_SUCC(node); \
  LNODE_PRED(LNODE_SUCC(node)) = LNODE_PRED(node); \
}

/* FEATURES */

LIST NewList(void)
{                                                 
  LIST result;
  
  result = (LIST) malloc(sizeof(struct list_s));
  if (result != NULL) {
    LIST_HEAD(result) = LIST_TAILER(result) ;
    LIST_TAIL(result) = LIST_HEADER(result) ;
    LIST_PRED(result) = NULL ;
    LIST_SIZE(result) = CSIZE_C(0) ;
  }
  return result;
}

void DelList(LIST list)
{
  ListFlush(list);
  free(list);
}

LNODE ListSearch(LIST list,CINDX idx)
{
  CINDX curidx;
  LNODE curnode;

  REQUIRE(IS_LIDX(list,idx));
    
  curidx = LIST_FAST_INDEX(list);
  curnode = LIST_FAST_LNODE(list);
  
  if (curidx < idx) {
    while (curidx < idx)
      {
	curidx++;
	curnode = LNODE_SUCC(curnode);
      }
  }
  else if (curidx > idx) {
    while (curidx > idx)
      {
	curidx--;
	curnode = LNODE_PRED(curnode);
      }
  }
  return curnode;
}

ITEM ListItem(LIST list,CINDX idx)
{
  CINDX last;

  REQUIRE(! ListIsEmpty(list));
  
  last = ListLastIndex(list);
  if (idx <= CINDX_C(0)) {
    LIST_FAST_LNODE(list) = LIST_HEAD(list);
    LIST_FAST_INDEX(list) = CINDX_C(0);
  }
  else if (idx >= last) { 
    LIST_FAST_LNODE(list) = LIST_TAIL(list);
    LIST_FAST_INDEX(list) = last;
  }
  else if (idx != LIST_FAST_INDEX(list)) {
    LIST_FAST_LNODE(list) = ListSearch(list,idx) ;
    LIST_FAST_INDEX(list) = idx;
  }
  return LNODE_ITEM(LIST_FAST_LNODE(list));
}

CINDX ListFind(LIST list,ITEM item)
{
  CINDX i;
  CINDX count;
  
  REQUIRE(IS_LIST(list));
  
  count = LIST_SIZE(list);
  for (i = CINDX_C(0); i < count; i++)
    if (ListItem(list,i) == item)
      return i;
  return CINDX_OUT;
}

BOOL ListPut(LIST list,ITEM item,CINDX idx)
{
  LNODE node;
  LNODE left;
  LNODE right;
  CINDX count;

  REQUIRE(IS_LIST(list));

  /* Create a new node */
  NEW_LNODE(node,item);

  count = LIST_SIZE(list);
  if (idx <= CINDX_C(0)) {
    idx = CINDX_C(0);
    left = LIST_HEADER(list);
  }
  else if (idx >= count) {
    idx = count;
    left = LIST_TAIL(list);
  }
  else
    left = ListSearch(list,idx - CINDX_C(1));

  /* Find the node corresponding to index `idx'-1 and insert the new 
     node between this node and the one whose current index is `idx'. */
     
  right = LNODE_SUCC(left);
  LINK_LNODE(left,right,node);

  /* Update the fast-access information */
  if (count <= CSIZE_C(0))
    {
    /* The list was empty before the insertion and thus 
       had no suitable fast-access information available. 
       We initialize this information to the first index. */
      LIST_FAST_INDEX(list) = CINDX_C(0) ;
      LIST_FAST_LNODE(list) = node ;
    }
  else if (idx <= LIST_FAST_INDEX(list))
    /* The list was not empty before the insertion and thus
       had suitable fast-access information avalaible.
       Since we have inserted a new node in the part of the
       list which is situated on the left of the fast-index,
       the node which now corresponds to the fast-index is
       the predecessor of the old fast-node. */
    LIST_FAST_LNODE(list) = LNODE_PRED(LIST_FAST_LNODE(list));

  /* Increment the size of the list */
  LIST_SIZE(list)++ ;
  return TRUE ;
}

ITEM ListSet(LIST list,ITEM item,CINDX idx)
{
  LNODE node;
  ITEM old_item;

  REQUIRE(IS_LIST(list));

  old_item = ListItem(list,idx);
  node = LIST_FAST_LNODE(list);
  LNODE_ITEM(node) = item;
  return old_item ;
}

void ListRemove(LIST list,CINDX idx)
{
  LNODE node;
  CINDX last;
  CINDX fast;

  REQUIRE(! ListIsEmpty(list));
  
  last = ListLastIndex(list);  
  if (idx <= CINDX_C(0))  {
    idx = CINDX_C(0);
    node = LIST_HEAD(list) ;
  }
  else if (idx >= last) {
    idx = last;
    node = LIST_TAIL(list) ;
  }
  else 
    node = ListSearch(list,idx) ;
  
  fast = LIST_FAST_INDEX(list);
  if (idx <= fast) {
    if (fast < last)
      LIST_FAST_LNODE(list) = LNODE_SUCC(LIST_FAST_LNODE(list)) ;
    else {
      LIST_FAST_INDEX(list)--;
      if (idx == fast)
	LIST_FAST_LNODE(list) = LNODE_PRED(LIST_FAST_LNODE(list)) ;
    }
  }

  UNLINK_LNODE(node) ;
  DEL_LNODE(node);

  LIST_SIZE(list)-- ;
}

void ListRemItem(LIST list,ITEM item)
{
  CINDX i;

  i = ListFind(list,item);
  if (i != CINDX_OUT)
    ListRemove(list,i);
}

void ListFlush(LIST list)
{
  REQUIRE(IS_LIST(list));
  
  while (! ListIsEmpty(list)) ListRemFirst(list);

  ENSURE(ListIsEmpty(list));
}

void ListSwap(LIST list,CINDX idx1,CINDX idx2)
{
  ITEM item1;
  ITEM item2;
  LNODE node1;
  LNODE node2;

  REQUIRE(IS_LIST(list));
  
  item1 = ListItem(list,idx1); 
  node1 = LIST_FAST_LNODE(list);  
  item2 = ListItem(list,idx2);
  node2 = LIST_FAST_LNODE(list);
  LNODE_ITEM(node1) = item2;
  LNODE_ITEM(node2) = item1;
}

void ListShiftL(LIST list)
{
  ITEM item;

  REQUIRE(IS_LIST(list));

  if (ListSize(list) < CSIZE_C(2)) return;
  item = ListFirstItem(list);
  ListRemFirst(list);
  ListPutLast(list,item);
}

void ListShiftR(LIST list)
{
  ITEM item;

  REQUIRE(IS_LIST(list));

  if (ListSize(list) < CSIZE_C(2)) return;
  item = ListLastItem(list);
  ListRemLast(list);
  ListPutFirst(list,item);
}

void ListPrint(LIST list,PFUNC print_item)
{
  CINDX i;
  CINDX count;

  REQUIRE(IS_LIST(list));

  printf("[");
  count = ListSize(list);
  if (count > CSIZE_C(0)) {
    print_item(ListItem(list,CINDX_C(0)));
    for (i = CINDX_C(1); i < count; i++)
      {
	printf(",");
	print_item(ListItem(list,i));
      }
  }
  printf("]");
}
