#include "checker.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* CHECKING FUNCTIONS */

void CheckStart(void)
{
  /* Reset check results before searching 
     CAUTION: the search time and number of iterations must not be reset
              here since we are already inside the search-loop */
  check_depth = UBINT_C(0);
  check_nbapp = UBINT_C(0);
  check_pmem = UBINT_C(0);
  check_kmem = UBINT_C(0);
}

void CheckSelect(void)
{
  /* We try to find an index allowing to end the proof-search directly.
     If not possible, let the method decide which index to start with. */

  if (! SeqIsBoxed()) {    
    /* The sequent must have a conclusion */
    switch(N_TY(N_CONCLU)) {
    case FTYPE_IMP:
      if (ContextHasIdx(N_RSF(N_CONCLU))) {
	select_fnidx = N_CONCLU;
	select_rule = r_imp;
	select_prem = PREM_FST;
	return;
      }
      break;
    case FTYPE_OR:
      if (ContextHasIdx(N_LSF(N_CONCLU))) {
	select_fnidx = N_CONCLU;
	select_rule = r_or;
	select_prem = PREM_FST;
	return;
      }
      if (ContextHasIdx(N_RSF(N_CONCLU))) {
	select_fnidx = N_CONCLU;
	select_rule = r_or;
	select_prem = PREM_SND;
	return;
      }
      break;
    case FTYPE_AND:
      if (ContextHasIdx(N_LSF(N_CONCLU))) {
	select_fnidx = N_CONCLU;
	select_rule = r_and;
	select_prem = PREM_FST;
	return;
      }
      if (ContextHasIdx(N_RSF(N_CONCLU))) {
	select_fnidx = N_CONCLU;
	select_rule = r_and;
	select_prem = PREM_SND;
	return;
      }
      break;
    case FTYPE_CST:
    case FTYPE_VAR:
      break;
    }
  }
  /* We failed to find an index allowing to end the derivation in one step.
     So we let the method decide which index to continue with. */
  select_fnidx = FINDX_OUT;
  method->Select();
}

void CheckExpand(void) 
{ 
  CHECK(select_fnidx != FINDX_OUT);
  CHECK(RuleIsPrem(select_rule,select_prem));

  /* Keep track of the first selected index of the current sequent */
  if (PE_START(Top()) == FINDX_OUT)
    PE_START(Top()) = select_fnidx;
  /* Add one visit to the selected index */
  N_PASS(select_fnidx)++; 
  check_nbapp++;
}

void CheckExpandSucc(void)
{
  if (PsSize() > check_depth) check_depth = PsSize();
  PE_PMEM(Top()) = UBINT_C(1); /* One node proof-tree */
  PE_KMEM(Top()) = UBINT_C(0); /* No kripke-tree */
}

void CheckExpandFail(void)
{
  if (PsSize() > check_depth) check_depth = PsSize();
  PE_PMEM(Top()) = UBINT_C(0); /* No proof-tree */
  PE_KMEM(Top()) = UBINT_C(1); /* One node kripke-tree */
}

void CheckPremissSucc(void)
{
  /* We must decide whether we can backtrack one step further or not. */
  if (RuleIsConj(RevRule()) && (N_PASS(RevIdx()) == UINT8_C(1))) { 
    /* The premiss is part of a conjunctive rule the other part of which
       has been tested yet. Therefore, we must try whether the other part
       (premiss) yields a proof or not. 
    */    
    select_fnidx = RevIdx(); /* Re-expand the same index */
    select_rule = RevRule(); /* Using the same rule */
    select_prem = PremSwitch(RevPrem()); /* With the other premiss */
  } else {
    /* Three possible cases:
       - the premiss is part of a single-premiss rule or,
       - the premiss is part of a disjunctive rule
       - the premiss is part of a conjunctive rule the other part of which
         has already been tested.
       In any case, the sequent is provable and we can backtrack further.
    */ 
    N_PASS(RevIdx()) = UINT8_C(0); /* Reset index pass-count */
    select_fnidx = FINDX_OUT; /* Continue backtracking */
  }
  /* Compute the size of the local proof-tree. If a proof-size has already 
     been computed for a previous premiss (no single-premiss rule), we must 
     add it to the proof-size returned by the reversed premiss.
     Otherwise, we must count one proof-node, for the reversed index, to 
     which we add the proof-size returned by the reversed premiss.
  */
  if (PE_PMEM(Top()) == UBINT_C(0))
    PE_PMEM(Top()) = RevPMem() + UBINT_C(1);
  else
    PE_PMEM(Top()) += RevPMem();
}

void CheckPremissFail(void)
{
  /* We must decide whether we can backtrack one step further or not. */
  if (RuleIsInver(RevRule(),RevPrem())) {
    /* The premiss is invertible, therefore, the counter-model it brings 
       back is also a counter-model to the current sequent. 
    */
    N_PASS(RevIdx()) = UINT8_C(0); /* Reset index pass-count */
    PE_KMEM(Top()) += RevKMem(); /* Estimate counter-model space size */
    select_fnidx = FINDX_OUT; /* Continue backtracking */
  } 
  else {
    if (RuleIsDisj(RevRule()) && (N_PASS(RevIdx()) == UINT8_C(1))) {
      /* The premiss is not invertible and is part of a disjunctive rule
	 the other part of which has not been tested yet. Therefore, we 
	 must try whether the other part (premiss) yields a proof or not. 
      */
      select_fnidx = RevIdx(); /* Re-expand the same index */
      select_rule = RevRule(); /* Using the same rule */
      select_prem = PremSwitch(RevPrem()); /* With the other premiss */
    } 
    else {
      /* The premiss is not invertible and
	 - is part of a single-premiss rule or,
	 - is part of a conjunctive rule or,
	 - is part of a disjunctive rule the other part of which has 
         already been tested.
	 In any case, we are done with the reversed index but, since the rule
	 is not invertible, we cannot immediately conclude that the current
	 sequent is unprovable. Indeed, we must try to find another untreated 
	 index to expand. If not possible, then, the sequent is unprovable.
      */ 
      N_PASS(RevIdx()) = UINT8_C(0); /* Reset index pass-count */
      method->Retry(); /* Try another choice (index,rule,premiss) */
    }
    /* Estimate counter-model space size */
    if (PE_KMEM(Top()) == UBINT_C(0))
      PE_KMEM(Top()) = RevKMem() + 1;
    else
      PE_KMEM(Top()) += RevKMem();
  }
  PE_PMEM(Top()) = UBINT_C(0); /* Reset proof-size information  */
}

void CheckEnd(void)
{
  /* Move search-stack informations to the corresponding global variables. */
  check_depth--; /* -1 because of the initial stack element */
  check_mbase = method_base;
  check_movhd = method_ovhd;
  check_fbase = CheckSize() * UBINT_C(2);
  check_fovhd = forimp_ovhd;
}

void CheckSuccess(void)
{
  check_pmem = TopPMem(); /* Retrieve proof-size */ 
  check_kmem = UBINT_C(0);
  CheckEnd();

  ENSURE(CheckPMem() > UBINT_C(0));
  ENSURE(CheckKMem() == UBINT_C(0));
}

void CheckFailure(void)
{
  check_kmem = TopKMem(); /* Retrieve countermodel-size */
  check_pmem = UBINT_C(0);
  CheckEnd();

  ENSURE(CheckPMem() == UBINT_C(0));
  ENSURE(CheckKMem() > UBINT_C(0));
}

