#include "forest.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* HELPERS */

#define ADDCLASS(t,c) ((t==FINDX_OUT) ? ((t)=(c)):(C_NX(c)=(t),(t)=(c)))

void BuildClass(FINDX n)
     /* Remark:
	n must not be the overall node. */
{
  FINDX c;
  PFORM f;

  f = N_FO(n);
  /* If some classes already exist, we check 
     whether n belongs to one of them */  
  for (c = FINDX_C(0); c < N_CCOUNT; c++)
    {
      if (FormIsEqu(f,C_FO(c)))
	{
	  /* n belongs to class c */
	  N_CL(n) = c;
	  N_NC(n) = C_FN(c);
	  C_FN(c) = n;
	  if (IdxIsRoot(n))
	    EnterContext(n);
	  return;
	}
    }
  /* First time we meet formula of node n in the forest */

  /* Build a new class where n is the only one element */
  c = N_CCOUNT++;
  C_FN(c) = n;
  C_IC(c) = IdxIsRoot(n) ? UBINT_C(1):UBINT_C(0);
  C_NX(c) = FINDX_OUT;
  C_DUP(c) = UBINT_C(0);
  N_CL(n) = c;
  N_NC(n) = FINDX_OUT;

  switch(FormType(f))
    {
    case FTYPE_CST: 
      if (FormIdxIsBot(f))
	N_CLSBOT = c;
      else 
	{
	  N_CLSTOP = c;
	  /* We have a class for the top unit. So, top belongs to the lexic 
	     of the sequent that we will have to decide. Therefore, we may 
	     reach sequents where top appears as the goal and such sequents 
	     are axioms even if the top unit is not explicitly present in 
	     the leaves of the forest. This is because top is always 
	     implicitly present on the left-hand-side of a sequent. 
	     To have the axiom-test work properly, we must add this 
	     implicit occurence to the context. Increasing C_IC by one 
	     leads to a C_IC counter that will never (if there is no bug) 
	     reach zero anymore.
	  */
	  C_IC(c)++;
	}
      break;
    case FTYPE_VAR: ADDCLASS(N_CLSVAR,c); break;
    case FTYPE_AND: ADDCLASS(N_CLSAND,c); break;
    case FTYPE_OR:  ADDCLASS(N_CLSOR,c);  break;
    case FTYPE_IMP: ADDCLASS(N_CLSIMP,c); break;
    }
}

void BuildClasses(void)
     /* Remark:
	the forest is supposed (and must be) indexed */
{
 FINDX n;

 N_CCOUNT = FINDX_C(0);
 N_CLSBOT = FINDX_OUT;
 N_CLSTOP = FINDX_OUT;
 N_CLSVAR = FINDX_OUT;
 N_CLSAND = FINDX_OUT;
 N_CLSOR =  FINDX_OUT;
 N_CLSIMP = FINDX_OUT;

 /* We compute the class of each node (except the overall node) */
 for (n = FINDX_C(1); n < N_NCOUNT; n++)
   BuildClass(n);
}

void FeedFormula(PFORM f)
/* FeedFormula(f) links each subformula of f (f included)
   to its corresponding node (index) in the forest area 
   Remark : 
     f is supposed (and must be) indexed
*/ 
{
  FINDX i;

  REQUIRE(IS_FORM(f)); 

  i = FormIndex(f);
  N_FO(i) = f;
  N_TY(i) = FormType(f);
  if (FormIsBinary(f))
    {
      N_LSF(i) = FormLSFIndex(f);
      N_RSF(i) = FormRSFIndex(f);
      FeedFormula(FormLSF(f));
      FeedFormula(FormRSF(f));
    }
}

/* CONSTRUCTOR */

void MakeForest(LFORM l,PFORM f)
/* Allocate memory for the sequent l |- f and indexes all the formulae 
   in l in the order of the list starting from index 1. The succedent 
   formula f is then indexed. Finally, the forest is initialized as a 
   list of roots, with each root corresponding to a formulae in l 
*/
{
  FSIZE ncount;
  FINDX i,last,current;

  REQUIRE(IS_FORM(f));
  REQUIRE(! IS_FOREST());

  /* Index each formula in l starting from 1,
     Index the conclusion f statring form the last generated index
     Allocate an array for the forest (struct + nodes + classes)
  */
  ncount = IndexConcForm(f,IndexLForm(l,CINDX_C(1)));
  forest = (FOREST) ALLOCMEM(sizeof(struct forest_s) +  
			     ncount * (sizeof(struct fnidx_s) + 
				       sizeof(struct fncls_s)));
  if (forest == NULL)
    Abort("AllocForest","Cannot allocate forest");
  
  N_NCOUNT = ncount;
  N_TNODE = (fnidx_t *) (forest + 1);
  N_TNCLS = (fncls_t *) &(N_NODE(N_NCOUNT));

  /* All nodes (array cells) are initialized. */
  for (i = FINDX_C(0); i < N_NCOUNT; i++) 
    {
      N_FO(i) = NULL;
      N_TY(i) = FTYPE_OUT;
      N_LB(i) = N_RB(i) = FINDX_OUT;
      N_FA(i) = N_FS(i) = FINDX_OUT;
      N_RO(i) = FINDX_OUT;
      N_LML(i) = N_RML(i) = FINDX_OUT;
      N_LL(i) = N_RL(i) = FINDX_OUT;
      N_LSF(i) = N_RSF(i) = FINDX_OUT;
      N_NC(i) = FINDX_OUT;
      N_CL(i) = FINDX_OUT;
      N_PASS(i) = UINT8_C(0);
    }

  /* The conclusion is initialized */
  FeedFormula(f);
  N_CONCLU = FormIndex(f);
  N_TAGGED = FINDX_OUT; /* no starred formula initially */

  /* We start initializing the overall node */
  N_LB(N_ORIGIN) = N_ORIGIN;
  N_RB(N_ORIGIN) = N_ORIGIN;

  if (! LFormIsEmpty(l))
    {
      /* The elements of l are chained as the roots of the forest */
      N_FIRST_ROOT = FormIndex(LFormForm(l));
      last = N_FIRST_ROOT;
      do {
	FeedFormula(LFormForm(l));
        current = FormIndex(LFormForm(l));
	N_RB(last) = current;
	N_LB(current) = last;
        N_FA(current) = N_ORIGIN;
	N_RO(current) = current;
	N_LML(current) = N_RML(current) = current;
	N_RL(last) = current;
	N_LL(current) = last;
	last = current;
	l = LFormTail(l);
      } while (! LFormIsEmpty(l));
      
      /* The circle is now being closed .... */      
      N_LL(N_FIRST_ROOT) = FINDX_OUT;
      N_RL(last) = FINDX_OUT;
      N_LB(N_FIRST_ROOT) = last;
      N_RB(last) = N_FIRST_ROOT;
      
      /* We finish initializing the overall node */
      N_LML(N_ORIGIN) = N_FIRST_ROOT;
      N_RML(N_ORIGIN) = last;
    }
  
  /* The node-classes are built */
  BuildClasses();
  forest_lform = l;
  forest_concl = f;
}

/* ACCESSORS */

FINDX SeqNextIdx(FINDX fnidx) 
{
  REQUIRE(IS_FNODE(fnidx));

  if (IdxIsLastLeaf(fnidx))
    if (SeqIsBoxed())
      return N_FIRST_LEAF;
    else
      return N_CONCLU;
  else if (IdxIsConclu(fnidx))
    return SeqFirstIdx();
  else
    return N_RL(fnidx);
}

FINDX SeqPrevIdx(FINDX fnidx)
{
  REQUIRE(IS_FNODE(fnidx));

  if (IdxIsFirstLeaf(fnidx))
    if (SeqIsBoxed())
      return N_LAST_LEAF;
    else 
      return N_CONCLU;
  else if (IdxIsConclu(fnidx))
    if (ForestIsEmpty())
      return N_CONCLU;
    else
      return N_LAST_LEAF;
  else
    return N_LL(fnidx);
}

LFORM ForestVars(void)
     /* This functions builds a set which contains 
	all the propositional variables present in
	the leaf-list of the current forest. 
        Remark: 
	  variable formulae are not duplicated before
          being added to the list. */
{
  FINDX i;
  LFORM result;

  REQUIRE(IS_FOREST());

  result = LFORM_EMPTY;
  /* We're going to loop on the leaves */
  i = N_FIRST_LEAF;
  while (i != FINDX_OUT)
    {
      /* Variables are roots of type var */
      if (IdxIsRoot(i) && (N_TY(i)==FTYPE_VAR))
	/* Check whether the variables have already been recorded */
	if (! LFormHasPtr(result,C_FO(N_CL(i))))
	    /* Add class-witness of "i" in the set */
	    result = ConsLForm(result,C_FO(N_CL(i)));
      i = N_RL(i);
    }
  return result;
}

/* DISPLAY */

void PrintIdx(FINDX idx,BOOL withindex)
{
  ubint_t nbpar;

  REQUIRE(IS_FNODE(idx));
  
  if (withindex) 
    printf("%u:",idx); /* Print the index */

  PrintForm(N_FO(idx));

  if (IdxIsConclu(idx)) {
    return; /* The conclusion is always considered as a root 
	       (even when it is not the case in the forest) */ 
  }

  if (IdxIsTagged(idx)) {
    printf("*"); /* Print the star, if any */
  }

  for (nbpar = UBINT_C(0); (N_FA(idx) != FINDX_OUT) && (! IdxIsRoot(idx)); 
       nbpar++)
    {
      idx = N_FA(idx); /* Move up one link (implication) */
      printf("->("); 
      PrintForm(N_FO(idx));
    }
  
  for (; nbpar > UBINT_C(0); nbpar--)
      printf(")");
}

void PrintForest(BOOL withindex)
{
  FINDX l;

  REQUIRE(IS_FOREST());

  l = N_FIRST_LEAF;
  while (l != FINDX_OUT) 
    {
      PrintIdx(l,withindex);
      if (l == N_LAST_LEAF) 
	printf(" ");
      else
	printf(", ");
      l = N_RL(l);
    }
  printf("=> ");
  if (N_CONCLU == FINDX_OUT)
    printf("*");  /* No conclusion, the forest must be tagged */
  else
    PrintIdx(N_CONCLU,withindex);
}

void PrintSequent(FINDX hilite)
{
  FINDX i;
  BOOL printed;

  REQUIRE(IS_FOREST());

  printed = FALSE;
  i = N_FIRST_LEAF;
  if (i != FINDX_OUT) {
    if (i == hilite) printf("@");
    PrintIdx(i,FALSE);
    while (i != N_LAST_LEAF)
      {
	i = N_RL(i);
	printf(", ");
	if (i == hilite) printf("@");
	PrintIdx(i,FALSE);
      }
    printed = TRUE;
  }
    
  for (i = FINDX_C(0); i < N_CCOUNT; i++) {
    if (C_DUP(i) > UBINT_C(0)) {
      if (printed) printf(", ");
      PrintForm(C_FO(i));
      printed = TRUE;
    }
  }

  if (printed) printf(" ");
  printf("=> ");

  if (N_CONCLU == FINDX_OUT) printf("*");
  else PrintIdx(N_CONCLU,FALSE);
}








