#ifndef __FOREST_H
#define __FOREST_H

#include "listform.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef struct fncls_s fncls_t;
struct fncls_s
{
  FINDX fnidx; /* Index of a forest node belonging to the class */
  FINDX nextc; /* Index of the next class with same connective type */
  ubint_t intro; /* Number of left side introductions of the class */
  ubint_t dupli; 
};  

typedef struct fnidx_s fnidx_t;
struct fnidx_s
{
  PFORM form; /* Associated formula */
  FTYPE type; /* Optimisation: type of the associated formula */
  FINDX cls; /* Index of the class of the node */
  FINDX link[12]; /* Idx links */
  uint8_t pass; /* Number of logical rules applied to the node */
};

typedef struct forest_s forest_t, *FOREST;
struct forest_s
{
  fnidx_t *tnode; /* Array of forest nodes */
  fncls_t *tncls; /* The forest node-classes */ 
  FSIZE ncount; /* Number of nodes in the tnode */
  FSIZE ccount; /* Number of classes in the tncls */
  FINDX tagged; /* The node of the starred formula */
  FINDX conclu; /* The node of the conclusion */
  FINDX clsbot; /* The class of the bottom unit (0) */
  FINDX clstop; /* The class of the top unit (1) */
  FINDX clsvar; /* The first class of type variable */
  FINDX clsand; /* The first class of type and */
  FINDX clsor;  /* The first class of type or */
  FINDX clsimp; /* The first class of type imp */
};

/*************
 * INTERFACE *
 *************/

implem LFORM forest_lform;
implem PFORM forest_concl;
implem deferred FOREST forest; /* Current forest */

#define IS_FOREST() (forest != NULL)

#define N_TNODE   (forest->tnode)
#define N_NCOUNT  (forest->ncount) /* The size of the global forest */
#define N_TNCLS   (forest->tncls)
#define N_CCOUNT  (forest->ccount) /* Number of classes */ 
#define N_CONCLU  (forest->conclu) /* Index of the conclusion */
#define N_TAGGED  (forest->tagged) /* Index of the starred formula */
#define N_CLSBOT  (forest->clsbot) /* Index of the class of falsity */
#define N_CLSTOP  (forest->clstop) /* Index of the class of truth */
#define N_CLSVAR  (forest->clsvar) /* Index of the first class of type var */
#define N_CLSAND  (forest->clsand) /* Index of the first class of type and */
#define N_CLSOR   (forest->clsor)  /* Index of the first class of type or */
#define N_CLSIMP  (forest->clsimp) /* Index of the first class of type imp */
#define N_GOAL    ((N_TAGGED == FINDX_OUT) ? N_CONCLU:N_TAGGED)

#define N_NODE(p) (N_TNODE[p]) /* Access to node whose index is p */ 
#define N_FO(p)   (N_NODE(p).form) /* The associated formula */
#define N_TY(p)   (N_NODE(p).type) /* Type of the formula */
#define N_LB(p)   (N_NODE(p).link[0]) /* Left brother */
#define N_RB(p)   (N_NODE(p).link[1]) /* Right brother */
#define N_FS(p)   (N_NODE(p).link[2]) /* First son */
#define N_FA(p)   (N_NODE(p).link[3]) /* Father */
#define N_RO(p)   (N_NODE(p).link[4]) /* Root */
#define N_LML(p)  (N_NODE(p).link[5]) /* Leftmost leaf beneath p */
#define N_RML(p)  (N_NODE(p).link[6]) /* Rightmost leaf beneath p  */
#define N_LL(p)   (N_NODE(p).link[7]) /* Left hand side leaf */
#define N_RL(p)   (N_NODE(p).link[8]) /* Right hand side leaf */
#define N_LSF(p)  (N_NODE(p).link[9]) /* Idx of the left subformula */
#define N_RSF(p)  (N_NODE(p).link[10]) /* Idx of the right subformla */
#define N_NC(p)   (N_NODE(p).link[11]) /* Next node in the class of p */
#define N_CL(p)   (N_NODE(p).cls) /* Class of node p */
#define N_PASS(p) (N_NODE(p).pass) /* How many premises were applied to p */
#define N_LS(p)   N_LB(N_FS(p)) /* Index of the last son of p */
#define N_SC(p)   FormScope(N_FO(p)) /* The scope of index p */

#define N_ORIGIN CINDX_C(0) /* Index of the overall node.
			       CAUTION: don't change this since many features
			       are implemented relying on this value to be 0.
			    */ 

#define N_FIRST_ROOT N_FS(N_ORIGIN) /* Index of the first root */
#define N_LAST_ROOT  N_LS(N_ORIGIN)  /* Index of the last root */
#define N_FIRST_LEAF N_LML(N_ORIGIN) /* Index of the first leaf */
#define N_LAST_LEAF  N_RML(N_ORIGIN) /* Index of the last leaf */

#define N_NCLS(n) (N_TNCLS[n]) /* Access to class of index n */
#define C_FN(c)   N_NCLS(c).fnidx /* Index of a forest node in class c */
#define C_IC(c)   N_NCLS(c).intro /* Number of LHS introductions of class c */ 
#define C_NX(c)   N_NCLS(c).nextc /* Index of the next formula in class c */
#define C_DUP(c)  N_NCLS(c).dupli
#define C_FO(c)   N_FO(C_FN(c)) /* The first formula of node-class c */

#define NC_IC(p)   C_IC(N_CL(p)) /* Occurences of p in the context */
#define NC_DUP(p)  C_DUP(N_CL(p))

/* ASSERTIONS */

#define IS_FNODE(n) (IS_FOREST() && (N_ORIGIN < (n)) && ((n) <= N_NCOUNT))

/* CONSTRUCTORS - DESTRUCTORS */

implem void MakeForest(LFORM l,PFORM p); 

#define KillForest() FREEMEM(forest)
			    
/* FOREST FEATURES */

#define ForestFirstRoot() GUARD(IS_FOREST(),N_FIRST_ROOT)
/* Return: the index of the first root of the forest or
           FINDX_OUT if there is no root in the forest */

#define ForestLastRoot()  GUARD(IS_FOREST(),N_LAST_ROOT)
/* Return: the index of the last root of the forest or
           FINDX_OUT if there is no root in the forest */

#define ForestFirstLeaf() GUARD(IS_FOREST(),N_FIRST_LEAF)
/* Return: the index of the first leaf of the forest or
           FINDX_OUT if there is no leaf in the forest */

#define ForestLastLeaf()  GUARD(IS_FOREST(),N_LAST_LEAF)
/* Return: the index of the last leaf of the forest or
           FINDX_OUT if there is no leaf in the forest */

#define ForestIsEmpty()  (ForestFirstRoot() == FINDX_OUT)
/* Return: TRUE if the forest has no active leaves */

#define ForestIsTagged() (N_TAGGED != FINDX_OUT) 
/* Return: TRUE if the forest has a starred leaf */

implem LFORM ForestVars(void);
/* Return: the list of propositional variables in the forest leaves */ 



/* FOREST NODE FEATURES */

#define IdxIsFirstRoot(p) (ForestFirstRoot() == (p)) 

#define IdxIsLastRoot(p)  (ForestLastRoot() == (p))  

#define IdxIsFirstLeaf(p) (ForestFirstLeaf() == (p)) 

#define IdxIsLastLeaf(p)  (ForestLastLeaf() == (p))  

#define IdxIsRoot(p) GUARD(IS_FNODE(p),N_FA(p) == N_ORIGIN) 

#define IdxIsLeaf(p) GUARD(IS_FNODE(p),N_FS(p) == FINDX_OUT) 

#define IdxIsAlone(p) GUARD(IS_FNODE(p),N_LB(p) == (p))  

#define IdxIsFirstSon(p) GUARD(IS_FNODE(p),N_FS(N_FA(p)) == (p))  

#define IdxIsLastSon(p) GUARD(IS_FNODE(p),IdxIsFirstSon(N_RB(p)))  

#define IdxIsTagged(p) GUARD(IS_FNODE(p),(p) == N_TAGGED) 

#define IdxIsConclu(p) GUARD(IS_FNODE(p),(p) == N_CONCLU) 

#define IdxIsGoal(p) GUARD(IS_FNODE(p),(p) == N_GOAL) 
/* Return: TRUE is 'p' is the conclusion or the tagged leaf */

#define IdxIsEquCls(p,q) (N_CL(p) == N_CL(q)) 
/* Return: TRUE if indexes 'p' and 'q' are in the same class */ 

#define IdxIsBot(p) FormIsBot(N_FO(p))
/* Return: TRUE if index 'p' is in the bottom class */


/* CONTEXT FEATURES */

#define ContextHasClass(c)  (((c) != FINDX_OUT) && (C_IC(c) > 0)) 
/* Return: TRUE if class c is in the context */

#define ContextHasIdx(p) GUARD(IS_FNODE(p),ContextHasClass(N_CL(p))) 
/* Return: TRUE if the class of node p is in the context */

#define ContextHasGoal() ContextHasIdx(N_GOAL)
/* Return: TRUE if the class of the goal is in the context */

#define ContextHasBot()  ContextHasClass(N_CLSBOT) 
/* Return: TRUE if bottom is in the context */

#define EnterContext(p) (NC_IC(p)++)
/* Put node p in the sequent context once */

#define LeaveContext(p) (ContextHasIdx(p) ? (NC_IC(p)--):0) 
/* Remove node p from the context once */


/* SEQUENT FEATURES */

#define Antecedent() VALUE(forest_lform)
/* Return: the list of formulas which are in the LHS of the current-sequent */
 
#define Conclusion() VALUE(forest_concl)
/* Return: the formula which is the conclusion of the current-sequent */

#define SeqIsBoxed() ForestIsTagged()
/* Is it a boxed sequent */

#define SeqIsAxiom() (ContextHasBot() || ContextHasGoal()) 
/* Is it an axiom */

#define SeqFirstIdx() (ForestIsEmpty() ? N_CONCLU:N_FIRST_LEAF)
/* This function returns the first index of the sequent
   Return
     - first leaf if there are leaves
     - conclusion otherwise
*/

#define SeqLastIdx()  (ForestIsTagged() ? N_LAST_LEAF:N_CONCLU)
/* This function returns the last index of the sequent
   Return
     - last leaf if the sequent is boxed
     - conclusion otherwise
*/

public FINDX SeqNextIdx(FINDX index);
/* This function returns the index of the node which is on 
   the right-hand side of node 'index' in the sequent
   The sequent is considered circularly, i.e:
   NextIdx(conclusion) = first_leaf (if there are leaves)
                         = conclusion (otherwise)
   NextIdx(last_leaf)  = first_leaf (if boxed sequent)
                         = conclusion (otherwise)
   Require: index != FINDX_OUT
*/

public FINDX SeqPrevIdx(FINDX index); 
/* This function returns the index of the node which is on the 
   left-hand side of node 'index' in the sequent
   The sequent is considered circularly, i.e:
   PrevIdx(first_leaf) = last_leaf (if boxed sequent)
                         = conclusion (otherwise)
   PrevIdx(conclusion) = last_leaf (if there are leaves)
                         = conclusion (otherwise)
   Require: index != FINDX_OUT
*/


/* DISPLAY */

public void PrintIdx(FINDX idx,BOOL withindex);
/* Print the implicational formula associated with an index, i.e., the 
   formula obtained by considering as implications all the links crossed
   in the upward path from 'idx'  to the root above 'idx'. More precisely,
   given the upward path [I0 ... IN], where I0 = 'idx' and IN is the root,
   the corresponding formula is F0->(F1->(...->(FN)...)), where Fk is the
   formula attached to index 'Ik'.
   Parameters
     'withindex' indicates whether or not the formula should be printed 
     prefixed by its index 'idx'.
*/

public void PrintForest(BOOL withindex);
/* Print the forest as a sequent, considering only the active formulae, i.e.,
   those formulae that are chained in the list of leaves. 
   Parameters
     'withindex' indicates whether or not the formulae should be printed 
     prefixed by their index.
*/

public void PrintSequent(FINDX hilite);
/* Print the forest as a sequent, considering all formulae present in 
   the context. 
   Parameters
     'withindex' indicates whether or not the formulae should be printed 
     prefixed by their index.
*/

#endif

