#include "formula.h"
#include "LIB/fifo.h"

/******************
 * IMPLEMENTATION *
 ******************/

const char *BotValue = "F";
const char *TopValue = "T";
const char *AndValue = "&";
const char *OrValue = "|";
const char *ImpValue = "->";
const char *NotValue = "~";

/* CONSTRUCTORS */

PFORM NewForm(FTYPE type,const char *val,PFORM lsf,PFORM rsf)
{       
  PFORM r;

  REQUIRE(IS_FTYPE_ATM(type) || 
	(IS_FTYPE_BIN(type) && (lsf != NULL) && (rsf != NULL) && (lsf!=rsf)));

  r = (PFORM) malloc(sizeof(struct formula_s));
  if (r == NULL)
    Abort("AllocForm","Cannot allocate formula");

  F_TY(r) = type;
  F_VA(r) = val; 
  F_LE(r) = lsf;
  F_RI(r) = rsf;
  F_SI(r) = FSIZE_C(1);
  if (IS_FTYPE_BIN(type))
    F_SI(r) = F_SI(r) + FormSize(lsf) + FormSize(rsf);
  F_IN(r) = FINDX_OUT;
  F_SC(r) = FINDX_OUT;
  return r;
}

PFORM DupForm(PFORM p)
{     
  REQUIRE(IS_FORM(p));

  if (FormIsAtomic(p))
    return NewForm(FormType(p),FormValue(p),NULL,NULL);
  else
    return NewForm(FormType(p),FormValue(p),
		   DupForm(FormLSF(p)),DupForm(FormRSF(p)));
}

/* DESTRUCTORS */

void DelForm(PFORM p)
{       
  if (! IS_FORM(p)) 
    return;
  if (FormIsBinary(p)) {
    DelForm(FormLSF(p));
    DelForm(FormRSF(p));
  } 
 free(p);
}

/* ACCESSORS */

BOOL FormIsEqu(PFORM p,PFORM q)
{  
  REQUIRE(IS_FORM(p) && IS_FORM(q));

  if (FormSize(p) != FormSize(q))
    return FALSE; /* Not the same size */
  if (strcmp(FormValue(p),FormValue(q)))
    return FALSE; /* Not the same value */
  if (FormIsBinary(p)) {
    if (! FormIsEqu(FormLSF(p),FormLSF(q)))
      return FALSE; /* Not the same left subformula */
    if (! FormIsEqu(FormRSF(p),FormRSF(q)))
      return FALSE; /* Not the same right subformula */
  }
  return TRUE;
}

BOOL FormIsSub(PFORM p,PFORM q)
{
  REQUIRE(IS_FORM(p) && IS_FORM(q));

  if (FormSize(p)>FormSize(q))
    return FALSE;
  if (FormIsEqu(p,q))
    return TRUE;
  if (FormIsBinary(q)) {
    if (FormIsSub(p,FormLSF(q)))
      return TRUE;
    if (FormIsSub(p,FormRSF(q)))
      return TRUE;
  }
  return FALSE;
}

/* DISPLAY */

void PrintForm(PFORM p)
{
  if (! IS_FORM(p))
    printf("null");
  else if (FormIsAtomic(p))
    printf("%s",FormValue(p));
  else if (FormIsImp(p) && FormIdxIsBot(FormRSF(p))) {
    printf("%s",NotValue);
    PrintForm(FormLSF(p));    
  }
  else {
    printf("(");
    PrintForm(FormLSF(p));
    printf("%s",FormValue(p));
    PrintForm(FormRSF(p));
    printf(")");    
  }
}

void PrintFormRaw(PFORM p)
{
  if (! IS_FORM(p))
    printf("null");
  else if (FormIsAtomic(p))
    printf("%s",FormValue(p));
  else {
    printf("(");
    PrintFormRaw(FormLSF(p));
    printf("%s",FormValue(p));
    PrintFormRaw(FormRSF(p));
    printf(")");    
  }
}

/* MODIFIERS */

FIFO  fifo;
FINDX next_id;

#define PUTFIFO(x)  FifoPush(fifo,(void *) (x))
#define GETFIFO(x)  (x = FifoPop(fifo))
#define EMPTYFIFO   FifoIsEmpty(fifo)

void IndexLeafRec(PFORM f)
{
  F_IN(f) = next_id++;
  switch (FormType(f))
    {
    case FTYPE_OR: 
      IndexLeafRec(FormLSF(f)); 
      F_SC(FormLSF(f)) = next_id - FINDX_C(1);
      IndexLeafRec(FormRSF(f)); 
      F_SC(FormRSF(f)) = next_id - FINDX_C(1);
      return;
    case FTYPE_AND:
      IndexLeafRec(FormRSF(f)); 
      IndexLeafRec(FormLSF(f));
      F_SC(FormLSF(f)) = next_id - FINDX_C(1);
      F_SC(FormRSF(f)) = next_id - FINDX_C(1);
      return;
    case FTYPE_IMP:
      PUTFIFO(FormLSF(f));
      IndexLeafRec(FormRSF(f));
      F_SC(FormRSF(f)) = next_id - FINDX_C(1);
      return;
    case FTYPE_CST:
    case FTYPE_VAR:
      return;
    }
}

void IndexRootRec(PFORM f)
{
  F_IN(f) = next_id++;
  switch (FormType(f))
    {
    case FTYPE_OR: 
    case FTYPE_AND:
      IndexRootRec(FormLSF(f));
      F_SC(FormLSF(f)) = next_id - FINDX_C(1);
      IndexRootRec(FormRSF(f));
      F_SC(FormRSF(f)) = next_id - FINDX_C(1);
      return;
    case FTYPE_IMP:
      IndexRootRec(FormRSF(f));
      IndexLeafRec(FormLSF(f));      
      F_SC(FormLSF(f)) = next_id - FINDX_C(1);
      F_SC(FormRSF(f)) = next_id - FINDX_C(1);
      return;
    case FTYPE_CST:
    case FTYPE_VAR:
      return;
    }
}

FINDX IndexLeftForm(PFORM f,FINDX start)
{
  REQUIRE(IS_FORM(f));

  fifo = NewFifo(FormSize(f));
  if (fifo == NULL)
    FatalError("IndexLeftForm","Cannot allocate fifo");

  next_id = start;
  PUTFIFO(f);
  do {
     GETFIFO(f);
     IndexRootRec(f);
     F_SC(f) = next_id - FINDX_C(1);
    }
  while (! EMPTYFIFO);

  DelFifo(fifo);
  return next_id;
}

void IndexConcRec(PFORM f)
{
  F_IN(f) = next_id++;
  F_SC(f) = FINDX_C(0);
  switch (FormType(f))
    {
    case FTYPE_OR: 
    case FTYPE_AND:
      IndexConcRec(FormLSF(f));
      IndexConcRec(FormRSF(f));
      return;
    case FTYPE_IMP:
      IndexLeftForm(FormLSF(f),next_id);
      IndexConcRec(FormRSF(f));
      return;
    case FTYPE_CST:
    case FTYPE_VAR:
      return;
    }
}

FINDX IndexConcForm(PFORM f,FINDX start)
{
  REQUIRE(IS_FORM(f));
  
  next_id = start;
  IndexConcRec(f);
  return next_id;
}
