#ifndef __FORMULA_H
#define __FORMULA_H

#include "LIB/platform.h"
#include "LIB/index.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* FORMULA SIZE */
typedef CSIZE FSIZE; /* Size of a collection */
#define FSIZE_MIN   CSIZE_MIN
#define FSIZE_MAX   CSIZE_MAX
#define FSIZE_C(c)  CSIZE_C(c)
#define IS_FSIZE(s) IS_CSIZE(s)

/* FORMULA INDEX */
typedef CINDX FINDX; /* Formula node index */
#define FINDX_MIN   CINDX_MIN
#define FINDX_MAX   CINDX_MAX
#define FINDX_OUT   CINDX_OUT
#define FINDX_C(c)  CINDX_C(c)
#define IS_FINDX(i) IS_CINDX(i)

/* FORMULA TYPES */
typedef enum { FTYPE_CST, FTYPE_VAR, FTYPE_AND, FTYPE_OR, FTYPE_IMP } FTYPE;
#define FTYPE_MIN FTYPE_CST
#define FTYPE_MAX FTYPE_IMP
#define FTYPE_OUT (FTYPE_MAX + 1)
#define IS_FTYPE(t) ((t) <= FTYPE_MAX)
#define IS_FTYPE_ATM(t) ((FTYPE_CST <= (t)) && ((t) <= FTYPE_VAR))
#define IS_FTYPE_BIN(t) ((FTYPE_AND <= (t)) && ((t) <= FTYPE_IMP))
#define FTYPE_TO_CINDX(t) ((FINDX) (t))

/* FORMULA NODE */

typedef struct formula_s *PFORM;
struct formula_s
{
  FTYPE f_type; /* Type of the formula */
  FSIZE f_size; /* Size of the formula (number of nodes) */
  FINDX f_index; /* Index of the formula */
  FINDX f_scope; /* Scope of the formula (subformula maximal index) */
  const char *f_value; /* Value of the formula */
  PFORM f_sub[2]; /* The left and right subformulae, if any */
};

/* Clean access to the fields of a formula-node */

#define F_TY(p) ((p)->f_type)
#define F_SI(p) ((p)->f_size)
#define F_IN(p) ((p)->f_index)
#define F_SC(p) ((p)->f_scope)
#define F_VA(p) ((p)->f_value)
#define F_LE(p) ((p)->f_sub[0])
#define F_RI(p) ((p)->f_sub[1])

/************
 * FEATURES *
 ************/

/* ASSERTIONS */

#define IS_FORM(p) ((p) != NULL)

/* LOGICAL SYMBOL VALUES */

public const char *BotValue;
public const char *TopValue;
public const char *AndValue;
public const char *OrValue;
public const char *ImpValue;
public const char *NotValue;

/* CONSTRUCTORS */

implem PFORM NewForm(FTYPE type,const char *val,PFORM lsf,PFORM rsf);

#define NewCst(val) NewForm(FTYPE_CST,(val),NULL,NULL)
/* Return: a new constant whose value is 'val' 
   Remark: 'val' is not duplicated
*/

#define NewVar(val) NewForm(FTYPE_VAR,(val),NULL,NULL)
/* Return: a new variable whose value is 'val' 
   Remark: 'val' is not duplicated
*/

#define NewAnd(p,q) NewForm(FTYPE_AND,AndValue,(p),(q))
/* Return: a new conjunction 'p' and 'q'
   Remark: the value of a conjunction is given by the feature AndValue
*/

#define NewOr(p,q)  NewForm(FTYPE_OR,OrValue,(p),(q))
/* Return: a new disjunction 'p' or 'q'
   Remark: the value of a disjunction is given by the feature OrValue
*/

#define NewImp(p,q) NewForm(FTYPE_IMP,ImpValue,(p),(q))
/* Return: a new implication 'p' implies 'q'
   Remark: the value of an implication is given by the feature ImpValue
*/

#define NewNot(p)   NewImp((p),NewCst(BotValue))
/* Return: a new implication 'p' implies bottom */

#define NewEqv(p,q) NewAnd(NewImp((p),(q)),NewImp(DupForm(q),DupForm(p)))
/* Return: a new biimplication ('p' implies 'q') and ('q' implies 'p') */


public PFORM DupForm(PFORM p);
/* Duplicate the formula 'p'
   Return: a new formula sharing no formula-nodes with 'p'.
   Require: IS_FORM(p)
   Remark: the values of the nodes are not duplicated
*/


/* DESTRUCTORS */

public void DelForm(PFORM p);
/* Destroy the formula 'p'
   Remark: nothing is done if 'p' is NULL */


/* ACCESSORS */

#define FormType(p) GETSAFE(IS_FORM(p),F_TY(p))
/* Return: the type of formula 'p' */

#define FormSize(p) GETSAFE(IS_FORM(p),F_SI(p))
/* Return: the size of formula 'p' */

#define FormIndex(p) GETSAFE(IS_FORM(p),F_IN(p))
/* Return: the index of formula 'p' */

#define FormScope(p) GETSAFE(IS_FORM(p),F_SC(p))
/* Return: the scope of formula 'p' */

#define FormValue(p) GETSAFE(IS_FORM(p),F_VA(p))
/* Return: the value of formula 'p' */


#define FormIsCst(p) (FormType(p) == FTYPE_CST)
/* Return: TRUE iff 'p' is a constant (top or bottom) */

#define FormIdxIsBot(p)  (FormIsCst(p) && (FormValue(p) == BotValue))
/* Return: TRUE iff 'p' is the constant "bottom" */

#define FormIsTop(p)  (FormIsCst(p) && (FormValue(p) == TopValue))
/* Return: TRUE iff 'p' is the constant "top" */

#define FormIsVar(p) (FormType(p) == FTYPE_VAR)
/* Return: TRUE iff 'p' is a propositional variable */

#define FormIsAtomic(p) IS_FTYPE_ATM(FormType(p))
/* Return: TRUE iff 'p' is atomic (variable or constant) */

#define FormIsAnd(p) (FormType(p) == FTYPE_AND)
/* Return: TRUE iff 'p' is a conjunction */

#define FormIsOr(p) (FormType(p) == FTYPE_OR)
/* Return: TRUE iff 'p' is a disjunction */

#define FormIsImp(p) (FormType(p) == FTYPE_IMP)
/* Return: TRUE iff 'p' is an implication */

#define FormIsBinary(p) IS_FTYPE_BIN(FormType(p))
/* Return: TRUE iff 'p' is a non-atomic formula */


#define FormLSF(p) GETSAFE(! FormIsAtomic(p),F_LE(p))
/* Return: the left subformula of 'p'
   Require: ! FormIsAtomic(p)
*/

#define FormRSF(p) GETSAFE(! FormIsAtomic(p),F_RI(p))
/* Return: the right subformula of 'p'
   Require: ! FormIsAtomic(p)
*/

#define FormLSFIndex(p) FormIndex(FormLSF(p))
/* Return: the index of the left subformula of 'p' 
   Require: ! FormIsAtomic(p)
*/

#define FormRSFIndex(p) FormIndex(FormRSF(p))
/* Return: the index of the right subformula of 'p' 
   Require: ! FormIsAtomic(p)
*/

#define FormLSFScope(p) FormScope(FormLSF(p))
/* Return: the scope of the left subformula of 'p' 
   Require: ! FormIsAtomic(p)
*/

#define FormRSFScope(p) FormScope(FormRSF(p))
/* Return: the scope of the right subformula of 'p' 
   Require: ! FormIsAtomic(p)
*/

public BOOL FormIsEqu(PFORM p,PFORM q);
/* Return: TRUE iff 'p' and 'q' are syntactically equivalent
   Require: IS_FORM(p) && IS_FORM(q)
*/

public BOOL FormIsSub(PFORM p,PFORM q);
/* Return: TRUE iff 'p' is syntactically contained in 'q' 
   Require: IS_FORM(p) && IS_FORM(q)
*/

/* DISPLAY */

public void PrintForm(PFORM p);
/* Print 'p' on stdout */

public void PrintFormRaw(PFORM p);
/* Print 'p' on stdout */


/* MODIFIERS */ 

public FINDX IndexLeftForm(PFORM p,FINDX start);

public FINDX IndexConcForm(PFORM p,FINDX start);



#endif
