#include "impisc.h"

void ISCExtendL(FINDX leaf, FINDX son)
/* This function links "son" as the first son of leaf "leaf"
   Remark:
     "leaf" might even be the overall node (empty forest) */
{
  REQUIRE(IS_FNODE(son));
  REQUIRE(IMP(leaf != N_ORIGIN,IS_FNODE(leaf) && IdxIsLeaf(leaf)));

  N_FS(leaf) = son;
  N_FA(son) = leaf;
  N_FS(son) = FINDX_OUT;
  N_LB(son) = N_RB(son) = son;
  if (leaf == N_ORIGIN)
    N_RO(son) = son;
  else
    N_RO(son) = N_RO(leaf);
  LinkL(N_LL(leaf),son);
  LinkL(son,N_RL(leaf));
}

void ISCRemoveL(FINDX leaf)
/* This function unlinks the leaf "leaf" out of the forest.
   Remarks:
     "leaf" might even be a root and is not modified.
     "leaf" must not be the overall node 
   Require:
     ! IdxIsFirst(leaf) */
{
  FINDX left,right,father;
  
  REQUIRE(IS_FNODE(leaf));
  REQUIRE(IdxIsLeaf(leaf));

  father = N_FA(leaf);
  if (IdxIsAlone(leaf)) {
    /* "leaf" has no brothers */
    /* disconnect "leaf" from its father */ 
    N_FS(father) = FINDX_OUT;
    if (IdxIsRoot(leaf)) {
      /* the forest is now empty */
      N_FIRST_LEAF = N_LAST_LEAF = FINDX_OUT;
    }
    else {
      /* replace 'leaf' by its parent */
      LinkL(N_LL(leaf),father);
      LinkL(father,N_RL(leaf));
    }
  }
  else {
    /* disconnect "leaf" from its bothers */
    left = N_LB(leaf);
    right = N_RB(leaf);
    N_LB(right) = left;
    N_RB(left) = right;
    /* disconnect "leaf" from the leaf list */
    LinkL(N_LL(leaf),N_RL(leaf));
    if (IdxIsFirstSon(leaf)) {
      /* disconnect "leaf" from its father */
      N_FS(father) = right;
    }
  }
}

void ISCReplaceL(FINDX leaf, FINDX new)
/* This function replaces the leaf "leaf" by the leaf "new"
   in the forest structure (modifying the connections).
   Remarks:
     "leaf" is not modified */
{
  REQUIRE(IS_FNODE(leaf));
  REQUIRE(IdxIsLeaf(leaf));

  N_FA(new) = N_FA(leaf); /* "new" will have the ame father as "leaf" */
  N_FS(new) = FINDX_OUT; /* "leaf" was a leaf and so must be "new" */

  if (IdxIsFirstSon(leaf)) /* if "leaf" was a first-son, so must be "new" */
    N_FS(N_FA(leaf)) = new;

  if (IdxIsRoot(leaf))
    N_RO(new) = new; /* "leaf" was a root-leaf, so must be "new" */
  else
    /* Link "new" with the same root as "leaf" when "leaf" is not a root */
    N_RO(new) = N_RO(leaf);

  if (IdxIsAlone(leaf)) {
    /* "leaf" had no brothers, so has "new" */
    N_LB(new) = N_RB(new) = new;
  }
  else {  
    /* Link "new" with the brothers of "leaf" */
    N_LB(new) = N_LB(leaf);
    N_RB(new) = N_RB(leaf);
    N_LB(N_RB(new)) = new;
    N_RB(N_LB(new)) = new;
  }

  /* Replace "leaf" by "new" in the leaf list */
  LinkL(N_LL(leaf),new);
  LinkL(new,N_RL(leaf));
}

void ISCNewRootL(FINDX leaf)
/* This function adds the leaf "leaf" as the last root of the forest. */
{
  FINDX last;

  REQUIRE(IS_FNODE(leaf));

  if (ForestIsEmpty()) {
    /* No root in the forest for the moment. */
    ISCExtendL(N_ORIGIN,leaf);
  }
  else {
    last = N_LAST_ROOT;
    N_FS(leaf) = FINDX_OUT;
    N_FA(leaf) = N_ORIGIN;
    N_RO(leaf) = leaf;
    N_RB(leaf) = N_RB(last);
    N_RB(last) = leaf;
    N_LB(N_RB(leaf)) = leaf;
    N_LB(leaf) = last;
    last = N_LAST_LEAF;
    LinkL(leaf,FINDX_OUT);
    LinkL(last,leaf);
  }
}

void ISCSplitLH(FINDX orig, FINDX left, FINDX right)
{
  REQUIRE(IS_FNODE(orig));

  N_FS(right) = FINDX_OUT;
  N_FA(right) = N_FA(orig);
  N_RB(right) = N_RB(orig);
  N_RB(orig) = right;
  N_LB(N_RB(right)) = right;
  N_LB(right) = orig;
  if (IdxIsRoot(orig))
    N_RO(right) = right;
  else
    N_RO(right) = N_RO(orig);
  LinkL(right,N_RL(orig));
  LinkL(orig,right);
  ISCReplaceL(orig,left);
}

void ISCUnsplitLH(FINDX orig, FINDX left, FINDX right)
{
  REQUIRE(IS_FNODE(orig));

  ISCRemoveL(right);
  ISCReplaceL(left,orig);
}

void ISCSplitLV(FINDX orig, FINDX up, FINDX down)
{
  REQUIRE(IS_FNODE(orig));

  ISCReplaceL(orig,up);
  ISCExtendL(up,down);
}

void ISCUnsplitLV(FINDX orig, FINDX up, FINDX down)
{
  REQUIRE(IS_FNODE(orig));

  ISCRemoveL(down);
  ISCReplaceL(up,orig);
}

void ISCDetachST(FINDX leaf,FINDX *fson,FINDX *lml,FINDX *rml)
/* This function removes the whole subtree under the father of
   leaf "leaf" (i.e "leaf" and all its brothers).
   Return:
     "fson" contains the index of the first son of the removed subtree
     "lml" contains the index of the left most leaf of the removed subtree
     "rml" contains the index of the right most leaf of the removed subtree
   Remark:
     "leaf" must not be the overall node
   Require:
     ! IdxIsFirst(leaf)
*/
{
  FINDX father,i;

  REQUIRE(IS_FNODE(leaf));

  father = N_FA(leaf);
  (*fson) = N_FS(father);

  /* Determine left most leaf */
  (*lml) = leaf;
  i = N_LL(leaf);
  while ((i>=father) && (i<=N_SC(father)))
    {
      forimp_ovhd++; /* one overhead move to the left */
      (*lml) = i;
      i = N_LL(i);
    }  
  /* Determine right most leaf */
  (*rml) = leaf;
  i = N_RL(leaf);
  while ((i>=father) && (i<=N_SC(father)))
    {
      forimp_ovhd++; /* one overhead move to the right */
      (*rml) = i;
      i = N_RL(i);
    } 

  LinkL(N_LL((*lml)),father);
  LinkL(father,N_RL((*rml)));
  N_FS(father)=FINDX_OUT;
}

void ISCAttachST(FINDX leaf, FINDX fson, FINDX lml, FINDX rml)
{
  FINDX father;

  REQUIRE(IS_FNODE(leaf));
  
  father = N_FA(leaf);
  N_FS(father) = fson;
  LinkL(N_LL(father),lml);
  LinkL(rml,N_RL(father));
}

void ISCInit(void) { } 
