#include "implrm.h"

/* HELPERS */

void LRMUpdateL(FINDX orig, FINDX left, FINDX right)
/* This function updates the information about the leftmost 
   and rightmost leaves in all the ancesters of "orig"
   setting them to "left" or "right" respectively.
   Remarks:
     "orig" is included and is therefore always modified. 
     The overall node might also be modified 
*/
{
  BOOL bl,br;

  REQUIRE(IS_FNODE(orig) || (orig == N_ORIGIN));

  /* Update of lrmost leaves for "orig" (constant part) */
  N_LML(orig) = left;
  N_RML(orig) = right;
  /* Update the information bottom-up (non constant part) */
  bl = br = TRUE;
  while ((bl || br) && (orig != N_ORIGIN))
    {
      bl = bl && IdxIsFirstSon(orig);
      br = br && IdxIsLastSon(orig);
      if (bl || br) {
	/* We must propagate the update upward */
	forimp_ovhd++; /* one update overhead */
	orig = N_FA(orig);
	if (bl) N_LML(orig) = left;
	if (br) N_RML(orig) = right;
      }
    }
}

/* FUNCTIONS */

void LRMExtendL(FINDX leaf, FINDX son)
/* This function links "son" as the first son of leaf "leaf"
   Remark:
     "leaf" might be even the overall node (empty forest) */
{
  REQUIRE(IS_FNODE(son));  
  REQUIRE(IMP(leaf != N_ORIGIN,IS_FNODE(leaf) && IdxIsLeaf(leaf)));

  N_FS(leaf) = son;
  N_FA(son) = leaf;
  N_FS(son) = FINDX_OUT;
  N_LB(son) = N_RB(son) = son;
  if (leaf == N_ORIGIN)
    N_RO(son) = son; /* "son" is a root since "leaf" is the origin */
  else
    N_RO(son) = N_RO(leaf); /* Chain "son" ot the root of "leaf" */
  LinkL(N_LL(leaf),son);
  LinkL(son,N_RL(leaf));
  LRMUpdateL(son,son,son);
}

void LRMRemoveL(FINDX leaf)
/* This function unlinks the leaf "leaf" out of the forest.
   Remarks:
     "leaf" might even be a root and is not modified.
     "leaf" must not be the overall node 
   Require:
     ! IdxIsFirst(leaf) */
{
  FINDX left,right,father;

  REQUIRE(IS_FNODE(leaf));
  REQUIRE(IdxIsLeaf(leaf));
  
  father = N_FA(leaf);
  if (IdxIsAlone(leaf)) {
    /* "leaf" has no brothers */
    /* disconnect "leaf" from its father */ 
    N_FS(father) = FINDX_OUT;
    /* update left and right most leaves */
    if (IdxIsRoot(leaf)) {
      /* the forest is now empty */
      N_FIRST_LEAF = N_LAST_LEAF = FINDX_OUT;
      LRMUpdateL(father,FINDX_OUT,FINDX_OUT);
    }
    else {
      LinkL(N_LL(leaf),father);
      LinkL(father,N_RL(leaf));
      LRMUpdateL(father,father,father);
    }
  }
  else {
    /* disconnect "leaf" from its bothers */
    left = N_LB(leaf);
    right = N_RB(leaf);
    N_LB(right) = left;
    N_RB(left) = right;
    
    /* disconnect "leaf" from the leaf list */
    LinkL(N_LL(leaf),N_RL(leaf));
    
    if (IdxIsFirstSon(leaf)) {
      /* disconnect "leaf" from its father */
      N_FS(father) = right;
      /* update left-most leaves */
      LRMUpdateL(father,N_LML(right),N_RML(father));
    }
    else if (IdxIsLastSon(leaf)) {
      /* update right-most leaves */
      LRMUpdateL(father,N_LML(father),N_RML(left));
    }
  }
}

void LRMReplaceL(FINDX leaf, FINDX new)
     /* This function replaces the leaf "leaf" by the leaf "new"
	in the forest structure (modifying the connections).
        Remarks:
	  "leaf" is not modified */
{
  REQUIRE(IS_FNODE(leaf));
  REQUIRE(IdxIsLeaf(leaf));

  N_FA(new) = N_FA(leaf);
  /* "Leaf" is assumed to be leaf, therefore so is "new" */
  N_FS(new) = FINDX_OUT;
  /* Link "new" to the father if "leaf" was the first son */
  if (IdxIsFirstSon(leaf))
    N_FS(N_FA(leaf)) = new;

  if (IdxIsRoot(leaf))
    /* "leaf" was a root-leaf, so is "new" */
    N_RO(new) = new;
  else
    /* Link "new" with the same root as "leaf" when "leaf" is not a root */
    N_RO(new) = N_RO(leaf);

  if (IdxIsAlone(leaf)) {
    /* "leaf" had no brothers, so has "new" */
    N_LB(new) = N_RB(new) = new;
  }
  else {  
    /* Link "new" with the brothers of "leaf" */
    N_LB(new) = N_LB(leaf);
    N_RB(new) = N_RB(leaf);
    N_LB(N_RB(new)) = new;
    N_RB(N_LB(new)) = new;
  }

  /* Replace "leaf" by "new" in the leaf list */
  LinkL(N_LL(leaf),new);
  LinkL(new,N_RL(leaf));
  LRMUpdateL(new,new,new);
}

void LRMNewRootL(FINDX leaf)
     /* This function adds the leaf "leaf" as the last root of the forest. */
{
  FINDX last;

  REQUIRE(IS_FNODE(leaf));

  if (ForestIsEmpty()) {
    /* No root in the forest for the moment. */
    LRMExtendL(N_ORIGIN,leaf);
  }
  else {
    last = N_LAST_ROOT;
    N_FS(leaf) = FINDX_OUT;
    N_FA(leaf) = N_ORIGIN;
    N_RO(leaf) = leaf;
    N_RB(leaf) = N_RB(last);
    N_RB(last) = leaf;
    N_LB(N_RB(leaf)) = leaf;
    N_LB(leaf) = last;
    last = N_LAST_LEAF;
    LinkL(leaf,FINDX_OUT);
    LinkL(last,leaf);
    LRMUpdateL(leaf,leaf,leaf);
  }
}

void LRMSplitLH(FINDX orig, FINDX left, FINDX right)
{
  REQUIRE(IS_FNODE(orig));

  N_FS(right) = FINDX_OUT;
  N_FA(right) = N_FA(orig);
  if (IdxIsRoot(orig))
    N_RO(right) = right;
  else
    N_RO(right) = N_RO(orig);
  N_RB(right) = N_RB(orig);
  N_RB(orig) = right;
  N_LB(N_RB(right)) = right;
  N_LB(right) = orig;
  LinkL(right,N_RL(orig));
  LinkL(orig,right);
  LRMReplaceL(orig,left);
  LRMUpdateL(right,right,right);
}

void LRMUnsplitLH(FINDX orig, FINDX left, FINDX right)
{
  REQUIRE(IS_FNODE(orig));

  LRMRemoveL(right);
  LRMReplaceL(left,orig);
}

void LRMSplitLV(FINDX orig, FINDX up, FINDX down)
{
  REQUIRE(IS_FNODE(orig));

  LRMReplaceL(orig,up);
  LRMExtendL(up,down);
}

void LRMUnsplitLV(FINDX orig, FINDX up, FINDX down)
{
  REQUIRE(IS_FNODE(orig));

  LRMRemoveL(down);
  LRMReplaceL(up,orig);
}

void LRMDetachST(FINDX leaf,FINDX *fson,FINDX *lml,FINDX *rml)
/* This function removes the whole subtree under the father of
   leaf "leaf" (i.e "leaf" and all its brothers).
   Return:
     "fson" contains the index of the first son of the removed subtree
     "lml" contains the index of the left most leaf of the removed subtree
     "rml" contains the index of the right most leaf of the removed subtree
   Remark:
     "leaf" must not be the overall node
   Require:
     !IdxIsFirst(leaf)
*/
{
  FINDX father;

  REQUIRE(IS_FNODE(leaf));

  father = N_FA(leaf);
  (*fson) = N_FS(father);
  (*lml) = N_LML(father);
  (*rml) = N_RML(father);
  LinkL(N_LL((*lml)),father);
  LinkL(father,N_RL((*rml)));
  N_FS(father) = FINDX_OUT;
  LRMUpdateL(father,father,father);
}

void LRMAttachST(FINDX leaf, FINDX fson, FINDX lml, FINDX rml)
{
  FINDX father;
  
  REQUIRE(IS_FNODE(leaf));

  father = N_FA(leaf);
  N_FS(father) = fson;
  LinkL(N_LL(father),lml);
  LinkL(rml,N_RL(father));
  LRMUpdateL(father,lml,rml);
}

void LRMInit(void) { }
