#include "kripke.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* KRIPKE-NODE MEMORY SPACE */

knode_t *ktree_node; /* Array of kripke-nodes */
KSIZE ktree_capa; /* Capacity of the array */
KSIZE ktree_used; /* Number of allocated nodes in the array */
KNODE ktree_free; /* Index of the first free node in the array */
CINDX ktree_lkid = CINDX_OUT; /* Last generated knode identifier */

/* ASSERTIONS */

#define IS_KNODE(n) (ktree_node && RANGE(KSIZE_C(0),(n),ktree_capa))

/* SET/GET MACROS FOR KRIPKE-NODES */
 
#define KT(n) ktree_node[(n)]
/* Kripke node 'n' */
#define KT_VA(n) (KT(n).forcing) 
/* The list of variables forced at node 'n' */
#define KT_FS(n) (KT(n).link[0]) 
/* The first son of node 'n' or 'n' itself if 'n' is a root */ 
#define KT_FA(n) (KT(n).link[1]) 
/* The father of node 'n' of 'n' itself if 'n' is a root */
#define KT_RB(n) (KT(n).link[2]) 
/* The right brother of node 'n' or KTREE_EMPTY if 'n' is a root */
#define KT_LB(n) (KT(n).link[3]) 
/* The left brother of node 'n' or KTREE_EMPTY if 'n' is a root */
#define KT_ID(n) (KT(n).knodeid)
/* The identifier of the node */

/* GET-ONLY MACROS FOR KRIPKE NODES */

#define KT_LS(n) KT_LB(KT_FS(n)) 
/* Return: the last son of node 'n' */

#define KT_ISROOT(n) GUARD(IS_KNODE(n),KT_FA(n) == (n))
/* Return: TRUE if node 'n' is a root ('n' points to a tree) */

#define KT_ISLEAF(n) GUARD(IS_KNODE(n),KT_FS(n) == (n))
/* Return: TRUE if node 'n' is a leaf ('n' points to a strict subtree) */

#define KT_ISFSON(n) GUARD(IS_KNODE(n),KT_FS(KT_FA(n)) == (n))
/* Return: TRUE if node 'n' is the first son of its father 
   Remark: FALSE if 'n' is a root */

#define KT_ISLSON(n) GUARD(IS_KNODE(n),KT_LS(KT_FA(n)) == (n))
/* Return: TRUE if node 'n' is the last son of its father
   Remark: FALSE if 'n' is a root */

#define KT_ISUSON(n) GUARD(IS_KNODE(n),KT_LB(n) == (n))
/* Return: TRUE if node 'n' is the unique son of its father
   Remark: FALSE if 'n' is a root */


/* CONSTRUCTORS */

BOOL MakeKSpace(KSIZE capacity)
{
  REQUIRE(ktree_node == NULL);
  REQUIRE(capacity > KSIZE_C(0));

  ktree_node = (knode_t *) malloc(sizeof(struct knode_s) * capacity);
  if (ktree_node == NULL) 
    return FALSE;
  /* Initialize memory space */
  ktree_capa = capacity; /* Set space capacity */
  ktree_used = KSIZE_C(0); /* No allocated node at the beginning */
  /* Chain free-nodes as a singly linked list */
  KT_LB(KNODE_C(0)) = KNODE_OUT;
  ktree_free = KNODE_C(0);
  while ( (ktree_free + KNODE_C(1)) < ktree_capa)
    {
      KT_LB(ktree_free + KNODE_C(1)) = ktree_free;
      ktree_free++;
    }
  return TRUE;
}

KTREE AllocKNode(void)
{
  KTREE result;

  REQUIRE(ktree_node != NULL);
  if (ktree_used >= ktree_capa)
    /* All nodes in the memory space have already been allocated */
    result = KTREE_EMPTY;
  else {
    result = ktree_free; /* Retrieve the first free-node in the space */
    ktree_free = KT_LB(result); /* Move the free ptr to the successor */
    ktree_used++; /* Count one more allocated node */
  }
  return result;
}

KTREE NewKTree(void)
{
  KTREE result;

  result = AllocKNode();
  if (result == KTREE_EMPTY)
    Abort("NewKTree","Cannot allocate memory for knode");
  else {
    KT_VA(result) = ForestVars();
    KT_FA(result) = result; /* to make first-son test work with a root-node */
    KT_FS(result) = result; /* to make last-son test work with a root-node */
    KT_RB(result) = KTREE_EMPTY; /* no right brother */
    KT_LB(result) = KTREE_EMPTY; /* no left brother */
    KT_ID(result) = CINDX_OUT;
  }
  return result;
}

/* DESTRUCTORS */

void KillKSpace(void)
{
  if (ktree_node != NULL) {
    free(ktree_node);
    ktree_node = NULL;
    ktree_free = KNODE_C(0);
    ktree_capa = KSIZE_C(0);
    ktree_used = KSIZE_C(0);
  }
}

void FreeKNode(KTREE k)
{
  REQUIRE(IS_KNODE(k));

  DelLForm(KT_VA(k),FALSE); /* Free formula-list without freeing formulae */
  ktree_used--; /* Count one less allocated node */
  KT_LB(k) = ktree_free; /* Chain the node as head of the free nodes list */
  ktree_free = k;
}

void DelKTreeRec(KTREE k)
{
  KTREE s1,s2;

  if (! KT_ISLEAF(k)) {
    /* We must delete all the sons of node k */
    s1 = KT_FS(k); /* Begin with the first son */
    do {
      s2 = KT_RB(s1); /* Save the right brother before deleting */
      DelKTreeRec(s1);
      s1 = s2;
    } while (! KT_ISFSON(s1));
  }
  /* Finally free kripke-node 'k' */ 
  FreeKNode(k); /* Free the node itself */
}

KTREE DelKTree(KTREE k)
{
  if (! KTreeIsEmpty(k)) {
    CHECK(KT_ISROOT(k)); /* k must be a root-node */
    DelKTreeRec(k);
  }
  return KTREE_EMPTY;
}

/* HELPERS */

void KTreeMapNodes(KTREE t,void (*mapfunc)(KNODE n));

void KTreeMapNodes(KTREE t,void (*mapfunc)(KNODE n))
{
  if (KTreeIsEmpty(t)) return;

 DESCEND:
  mapfunc(t);
  while (! KT_ISLEAF(t)) 
    {
      t = KT_FS(t);
      mapfunc(t);
    }
  while (! KT_ISROOT(t)) 
    {
      if (! KT_ISLSON(t)) {
	t = KT_RB(t);
	goto DESCEND;
      }
      t = KT_FA(t);
    }
}

void KTreeMapEdges(KTREE t,void (*mapfunc)(KNODE src,KNODE tgt));

void KTreeMapEdges(KTREE t,void (*mapfunc)(KNODE src,KNODE tgt))
{
  if (KTreeIsEmpty(t)) return;

 DESCEND:
  while (! KT_ISLEAF(t)) 
    {
      t = KT_FS(t);
      mapfunc(KT_FA(t),t);
    }
  while (! KT_ISROOT(t)) 
    {
      if (! KT_ISLSON(t)) {
	t = KT_RB(t);
	mapfunc(KT_FA(t),t);
	goto DESCEND;
      }
      t = KT_FA(t);
    }
}

/* MODIFIERS */

KTREE AddKTree(KTREE t,KTREE s)
{
  KTREE first,last;

  if (KTreeIsEmpty(t)) return s;
  if (KTreeIsEmpty(s)) return t;
  if (KT_ISLEAF(t)) {
    /* t has no sons, s is added as its first son */
    KT_FS(t) = s; /* s becomes the first son of t */
    KT_FA(s) = t; /* t becomes the father of s */
    KT_LB(s) = s; /* s becomes its own left brother */ 
    KT_RB(s) = s; /* s becomes its own right brother */
  } else {
    /* t already has sons, s is added as its last son */
    KT_FA(s) = t; /* t becomes the father of s */
    last = KT_LS(t); /* get the last son of t */
    first = KT_FS(t); /* get the first son of t */
    KT_RB(s) = first; /* the first son becomes the right brother of s */
    KT_LB(s) = last; /* the last son becomes the left brother of s */ 
    KT_LB(first) = s; /* s becomes the left brother of the first son */
    KT_RB(last) = s; /* s becomes the right brother of the last son */
  }
  return t;
}

/* ACCESSORS */

KSIZE KTreeSize(KTREE k)
{
  KSIZE result;

  if (KTreeIsEmpty(k)) 
    return KSIZE_C(0);
  if (KT_ISLEAF(k))  
    return KSIZE_C(1);
  result = KSIZE_C(1);
  k = KT_FS(k);
  do {
    result += KTreeSize(k);
    k = KT_RB(k);
  } while (! KT_ISFSON(k));
  return result;
}

KSIZE KTreeDepth(KTREE k)
{
  KSIZE result,temp;

  if (KTreeIsEmpty(k)) 
    return KSIZE_C(0);
  if (KT_ISLEAF(k)) 
    return KSIZE_C(0);
  result = KSIZE_C(0);
  k = KT_FS(k);
  do {
    temp = KTreeDepth(k);
    if (temp > result)
      result = temp;
    k = KT_RB(k);
  } while (! KT_ISFSON(k));
  return result + KSIZE_C(1);
}

BOOL KTreeForceForm(KTREE k,FINDX i)
{
  switch(N_TY(i))
    {
    case FTYPE_CST: 
      return FormIsTop(N_FO(i)); 
    case FTYPE_VAR: 
      return LFormHasPtr(KT_VA(k),C_FO(N_CL(i))); 
    case FTYPE_AND: 
      return KTreeForceForm(k,N_LSF(i)) && KTreeForceForm(k,N_RSF(i)); 
    case FTYPE_OR:  
      return KTreeForceForm(k,N_LSF(i)) || KTreeForceForm(k,N_RSF(i));
    case FTYPE_IMP:
      if (KTreeForceForm(k,N_LSF(i)) && ! KTreeForceForm(k,N_RSF(i)))
	return FALSE;
      else if (KT_ISLEAF(k))
	return TRUE;   
      else {
	k = KT_FS(k);
	do {
	  if (! KTreeForceForm(k,i))
	    return FALSE;
	  k = KT_RB(k);
	} while (! KT_ISFSON(k));
      }
    }
  return TRUE;
}

BOOL KTreeForceIdxRec(KTREE k,FINDX i)
{
  if (IdxIsRoot(i))
    return KTreeForceForm(k,i);
  
  if (KTreeForceForm(k,i) && ! KTreeForceIdxRec(k,N_FA(i)))
    return FALSE;

  if (KT_ISLEAF(k))
    return TRUE;

  k = KT_FS(k);
  do {
    if (! KTreeForceIdxRec(k,i))
      return FALSE;
    k = KT_RB(k);
  } while (! KT_ISFSON(k));

  return TRUE;
}

BOOL KTreeForceIdx(KTREE k,FINDX i)
{
  REQUIRE(IS_FNODE(i));

  return KTreeIsEmpty(k) ? FALSE : KTreeForceIdxRec(k,i);
}

BOOL KTreeForceLHS(KTREE k)
{
  FINDX i;

  if (KTreeIsEmpty(k)) 
    return FALSE;

  i = N_FIRST_LEAF;
  while (i != FINDX_OUT)
    {
      if (! KTreeForceIdxRec(k,i))
	return FALSE;
      i = N_RL(i);
    }
  return TRUE;
}

BOOL KTreeForceRHS(KTREE k)
{
  return KTreeForceForm(k,N_GOAL);
}

/* INDEXING FEATURES */

void IndexKNode(KNODE n)
{
  REQUIRE(IS_KNODE(n));

  ktree_lkid++;
  KT_ID(n) = ktree_lkid;
}

void IndexKTree(KTREE t)
{
  if (KTreeIsEmpty(t)) return; /* No need to index an empty tree */
  if (KT_ID(t) != CINDX_OUT) return; /* The tree is already indexed */
  
  ktree_lkid = CINDX_OUT;
  KTreeMapNodes(t,IndexKNode);
}


/* DISPLAY FEATURES */

/* Screen printing routines */

void PrintKTreeRec(KTREE t,unsigned int step)
{
  /* We first print the node itself */
  PrintSpace(step);
  printf("%-2" PRIuCINDX ": { ",KT_ID(t));
  PrintLForm(KT_VA(t),PrintForm);
  printf(" }\n");
  /* Then we print all the sons (if any) */
  if (! KT_ISLEAF(t)) {
    t = KT_FS(t);
    do {
      PrintKTreeRec(t,step + 2U);
      t = KT_RB(t);
    } while (! KT_ISFSON(t));
  } 
}

void PrintKTree(KTREE t)
{
  if (! KTreeIsEmpty(t)) {
    IndexKTree(t);
    PrintKTreeRec(t,0U);
  }
}

/* GDL CONVERSION FEATURES */

void KTreeNodeToGDL(KNODE n)
{
  REQUIRE(IS_KNODE(n));

  printf("node: { title: \"%" PRIuCINDX "\" label: \"",KT_ID(n));
  PrintLForm(KT_VA(n),PrintForm);
  printf("\" }\n");
}

void KTreeEdgeToGDL(KNODE s,KNODE t)
{
  REQUIRE(IS_KNODE(s));
  REQUIRE(IS_KNODE(t));

  printf("edge: { " 
	 "sourcename: \"%" PRIuCINDX "\" " 
	 "targetname: \"%" PRIuCINDX "\" }\n", 
	 KT_ID(s), KT_ID(t));
}

void KTreeToGDL(KTREE t)
{
  printf("graph: {\n"
	 "label: \"");
  PrintSequent(FALSE);
  printf("\"\n");
  printf("layoutalgorithm: tree\n"
	 "orientation: bottom_to_top\n");
  IndexKTree(t);
  KTreeMapNodes(t,KTreeNodeToGDL);
  KTreeMapEdges(t,KTreeEdgeToGDL);
  printf("}\n");
}

/* XML CONVERSION FEATURES */

void KTreeNodeToXML(KNODE n)
{
  REQUIRE(IS_KNODE(n));

  printf("<NODE id=%" PRIuCINDX " lab=\"",KT_ID(n));
  PrintLForm(KT_VA(n),PrintForm);
  printf("\">\n");
}

void KTreeEdgeToXML(KNODE s,KNODE t)
{
  REQUIRE(IS_KNODE(s));
  REQUIRE(IS_KNODE(t));

  printf("<EDGE src=%" PRIuCINDX " tar=%" PRIuCINDX ">\n", KT_ID(s), KT_ID(t));
}

void KTreeToXML(KTREE t)
{
  printf("<GRAPH name=\"");
  PrintForest(FALSE);
  printf("\">\n");
  IndexKTree(t);
  printf("<NODES>\n");
  KTreeMapNodes(t,KTreeNodeToXML);
  printf("</NODES>\n");
  printf("<EDGES>\n");
  KTreeMapEdges(t,KTreeEdgeToXML);
  printf("</EDGES>\n");
  printf("</GRAPH>\n");
}

/* METAPOST CONVERSION FEATURES */

void KTreeNodeToMP(KNODE n)
{
  REQUIRE(IS_KNODE(n));

  printf("leaf.var%" PRIuCINDX "(btex %" PRIuCINDX " \\{ ",KT_ID(n),KT_ID(n));
  PrintLForm(KT_VA(n),PrintForm);
  printf(" \\} etex);\n");
}

void KTreeEdgeToMP(KNODE s,KNODE t)
{
  REQUIRE(IS_KNODE(s));
  REQUIRE(IS_KNODE(t));

  printf("branches.var%" PRIuCINDX "(var%" PRIuCINDX ");\n",KT_ID(s),KT_ID(t));
}

void KTreeToMP(KTREE t)
{
  printf("verbatimtex\n"
	 "\\documentclass{article}\n"
	 "\\usepackage[T1]{fontenc}"
	 "\\begin{document}\n"
	 "etex;\n"
	 "input marctree;\n"
	 "beginfig(0);\n"
	 "picture var[];\n");

  IndexKTree(t);
  KTreeMapNodes(t,KTreeNodeToMP);
  KTreeMapEdges(t,KTreeEdgeToMP);

  printf("draw treeit(var0);\n"
	 "endfig;\n"
	 "end\n");
}



