#ifndef __KRIPKE_H
#define __KRIPKE_H

#include "forest.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef CSIZE KSIZE;
#define KSIZE_MIN  CSIZE_MIN
#define KSIZE_MAX  CSIZE_MAX
#define KSIZE_C(c) CSIZE_C(c)

typedef CINDX KNODE;
#define KNODE_MIN  CINDX_MIN
#define KNODE_MAX  CINDX_MAX
#define KNODE_OUT  CINDX_OUT
#define KNODE_C(c) CINDX_C(c)

typedef KNODE KTREE;
#define KTREE_EMPTY KNODE_OUT

#define PRIuKNODE PRIuCINDX
#define SCNuKNODE SCNuCINDX

typedef struct knode_s knode_t;
struct knode_s
{
  LFORM forcing; /* List of propositional variables forced by the node */
  KNODE link[4]; /* Links to other nodes (common n-ary tree) */
  CINDX knodeid;
};

/*************
 * INTERFACE *
 *************/

/* CONSTRUCTORS - DESTRUCTORS */

public BOOL MakeKSpace(KSIZE capacity);
/* Return: TRUE if a enough memory space has been allocated to hold
   a kripke-tree containing at most 'capacity' node
   Require: capacity > 0 
*/

public KTREE NewKTree(void);
/* Return: a new kripke-node (one-node kripke-tree) in which all 
   propositional variables of the current forest are forced or
   KTREE_EMPTY if the memory space is full.
*/

public void KillKSpace(void);
/* Free the currently allocated memory space */

public KTREE DelKTree(KTREE k);
/* Destroy the kripke-tree 'k'
   Return: KTREE_EMPTY
   Require: KTreeIdxIsRoot(k)
   Remarks:
     All the memory allocated for 'k' is freed but nothing is done on links.
     Also, the formulae in the forcing list of each node is not freed.
*/


/* MODIFIERS */

public KTREE AddKTree(KTREE k,KTREE s);
/* Link kripke-tree 's' as a son of kripke-tree 't'.
   Remark: 
   - if 't' is empty, then, the result is 's'
   - if 's' is empty, then, the result is 't' (unchanged)
*/


/* ACCESSORS */

#define KTreeIsEmpty(k) ((k) == KTREE_EMPTY)
/* Return: TRUE if 'k' is an empty tree */

public KSIZE KTreeSize(KTREE k);
/* Return: the number of nodes in kripke-tree 'k' 
   Remark: the size of an empty kripke-tree is 0
*/

public KSIZE KTreeDepth(KTREE k);
/* Return: the number of edges of the longest branch of kripke-tree 'k'
   Remark: 
   - the depth of a kripke-node is 0.
   - the depth of an empty kripke-tree is 0.
*/

public BOOL KTreeForceIdx(KTREE k,FINDX i);
/* Return: TRUE if kripke-tree 'k' forces the formula associated with
   forest-node 'i', i.e., the formula built up moving from 'i' upward 
   to the root above 'i' considering each link as an implication.
   Require: i != FINDX_OUT
*/

public BOOL KTreeForceLHS(KTREE k);
/* Return: TRUE if kripke-tree 'k' forces the leaves of the current forest
*/

public BOOL KTreeForceRHS(KTREE k);
/* Return: TRUE if kripke-tree 'k' forces the conclusion of the current forest
*/

#define KTreeIsCM(k) (KTreeForceLHS(k) && (! KTreeForceRHS(k)))
/* Return: TRUE if kripke-tree 'k' is a counter-model for the current sequent 
*/


/* DISPLAY */

public void PrintKTree(KTREE k);

public void KTreeToMP(KTREE k);

public void KTreeToGDL(KTREE k);

public void KTreeToXML(KTREE k);

#endif

