/* This module defines a singly-linked list of formula-pointers */

#ifndef __LISTFORM_H
#define __LISTFORM_H

#include "formula.h"

/* TYPES */

/* Data-structure for a singly-linked list-node */

typedef	struct lform_s *LFORM;		
struct lform_s
{
  PFORM lf_form; /* Pointer to a formula */
  LFORM lf_tail; /* Pointer to the next node in the list */
};

/* Clean access to the fields of a list-node */

#define LF_FORM(l)  ((l)->lf_form)
#define LF_TAIL(l)  ((l)->lf_tail)

/*************
 * INTERFACE *
 *************/

/* CONSTRUCTORS - DESTRUCTORS */

#define LFORM_EMPTY NULL

public LFORM ConsLForm(LFORM l,PFORM p);
/* Create a new list-node which pointing to formula 'p'
   and links list 'l' as the tail of this new node.
   Return: 
     a list whose first node contains 'p' and whose tail is 'l'
   Remark: 
     to avoid pointer handling troubles, and unless you clearly know
     what you are doing, it is strongly recommended that you reaffect 
     to 'l' the result of the function, e.g : l = ConsLForm(l,p);
*/

public void DelLForm(LFORM l,BOOL freealsoform);
/* Delete list 'l' and frees the memory allocated for list-nodes. 
   Parameters:
    - 'l': the list to delete
    - 'freealsoform': if set to TRUE, the function also deletes the
    formulae referenced in the list-nodes.
*/

/* ACCESSORS */

#define LFormIsEmpty(l) ((l) == LFORM_EMPTY)
/* Return: TRUE iff 'l' has no elements */

#define LFormForm(l) GETSAFE(! LFormIsEmpty(l),LF_FORM(l))
/* Return: the formula referenced by list-node 'l'
   Require: ! LFormIsEmpty(l)
*/

#define LFormTail(l) GETSAFE( ! LFormIsEmpty(l),LF_TAIL(l))
/* Return: the tail (remainder) of list 'l'
   Require: ! LFormIsEmpty(l)
*/

public CSIZE LFormSize(LFORM l);
/* Return: the number of elements in list 'l' */

public BOOL LFormHasPtr(LFORM l,PFORM p);
/* Return: TRUE iff list 'l' has a list-node pointing to the formula 'p' */

public BOOL LFormHasSame(LFORM l,PFORM p);
/* Return: TRUE iff list 'l' contains a list-node which points
   to a formula which is syntactically equivalent to formula 'p' */

/* MODIFIERS */

typedef BOOL (*TMAPLFORM)(PFORM f);

public BOOL MapLForm(LFORM l,TMAPLFORM mapfunc);
/* Apply the function 'mapfunc' to list 'l'
   Parameters:
     - 'l': the list to apply 'mapfunc' to
     - 'mapfunc': the function to apply successively to each formula of
     the list 'l'; this function receives a formula of 'l' as parameter
     and should return TRUE if succeeds on this formula and FALSE if it
     fails. In that case, the mapping process stops and subsequent
     formulae of 'l' are not visited.
   Return:
     TRUE iff all formulae of the list have been successfully treated. 
*/

public LFORM DupLForm(LFORM l,BOOL dupalsoform);
/* Duplicate list 'l'
   Parameters:
    - 'l': the list to duplicate
    - 'dupalsoform': if set to TRUE, the function also duplicates the 
    formulae referenced in the list-nodes. In this case, the list-nodes 
    of the duplicated list will not point to the same formulae as those 
    in 'l', but to syntactically equivalent formulae.
   Return: 
     a list which is equivalent to 'l' and sharing no memory with 'l'
*/

public CINDX IndexLForm(LFORM l,CINDX start);
/* Index each formula in list 'l' 
   Parameters:
     - 'l': the list to be indexed
     - 'start': the value of the initial index
   Return: 
     the value of the last index used + 1
   Remark:
     Formulae in the left-hand-side of a sequent are not indexed in the 
     same way as the conclusion. A list of formulae is always considered 
     as a left-hand-side of a sequent even if it contains only one formula.
*/

/* DISPLAY */

public void PrintLForm(LFORM l,void (*pfunc)(PFORM p));
/* Write list 'l' on stdout using 'pfunc' as the printing function
   Parameters:
     - 'l': the list to write
     - 'pfunc': the printing function applied to each formula in 'l'
   Remark:
     formulae of 'l' are written separated from each-other by a comma 
*/

#endif
