/* This module implements the rule-precedence strategy */

#include "metrprec.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* FUNCTIONS */

void FindNextPrec(RPREC p,FINDX lo,FINDX hi)
/* Find the first expandable index, scanning left to right from index 'lo' 
   toward index 'hi', whose precedence is the highest to be lower than or 
   equal to 'p'. If no such index can be expanded in the specified range,
   the function sets the global variable 'select_fnidx' to FINDX_OUT.
   Otherwise, the selected triple (index,rule,premiss) is returned through 
   the global variables 'select_fnidx', 'select_rule', 'select_prem'.
   Remark
     Specifying a range in which 'lo' is higher than 'hi' corresponds to 
     [lo,last] U [first,hi], i.e., to a circular scanning of the range.
*/
{
  RULE rule;
  RPREC prec,select_prec;

  REQUIRE(IS_RPREC(p));
  REQUIRE(IS_FNODE(lo));
  REQUIRE(IS_FNODE(hi));
 
  method_base++; /* Count one irreducible look at index 'lo' */
  select_fnidx = FINDX_OUT; /* No selection at the beginning */
  select_prec = RPREC_OUT;
 
  while (TRUE) 
    {
      rule = FindRule(lo); /* Find if a rule applies to the current index */
      if (rule != RULE_OUT) {
	/* We have found an index to which we can apply a rule */
	prec = RulePrec(rule); /* Get the precedence of that rule */
	if (prec <= p) {
	  /* The precedence of the rule is less than or equal to 'p', thus,
	     the current index belongs to the set of selectable indexes */ 
	  if (select_fnidx == FINDX_OUT) {
	    /* No selection has been made so far and the current index
	       is selectable, so we select it. */
	    select_fnidx = lo;
	    select_rule = rule;
	    select_prec = prec;
	  }
	  else if (select_prec < prec) {
	    /* An index has been selected already but the current index 
	       has higher precedence. Thus, we select the current index. */
	    select_fnidx = lo;
	    select_rule = rule;
	    select_prec = prec;
	  }
	  else {
	    /* An index has been selected already and its precedence is
	       higher than or equal to the precedence of the current index.
	       So, the current index is either not the one with highest
	       suitable precedence, or not the first one to have it. */ 
	  }
	}
      }

      if ((lo == hi) || (select_prec == p))
	/* We stop either if the current index was the last we had to 
	   look at, or if the currently selected index has the highest 
	   suitable precedence, i.e., a precedence equal to 'p'. */
	break;

      /* Otherwise, we move to next index, counting one overhead move. */
      lo = SeqNextIdx(lo);
      method_ovhd++;
    }

  if (select_fnidx != FINDX_OUT) {
    /* Choose one premiss of the selected rule */
    if (RuleIsSing(select_rule) || RuleIsInver(select_rule,PREM_FST))
      /* The rule has only one premiss or the first premiss is invertible.
	 Thus, we choose the first premiss. */      
      select_prem = PREM_FST;
    else if (RuleIsInver(select_rule,PREM_SND))
      /* The rule has two premises, the first one is not invertible but the
	 second is, so, we choose the second premiss. */
      select_prem = PREM_SND;
    else
      /* The rule has to premises and both are not invertible, so we choose 
	 the first premiss. */
      select_prem = PREM_FST;
  }
}

void RPRECSelect(void)
{
  FindNextPrec(RPREC_MAX,SeqFirstIdx(),SeqLastIdx());
}

void RPRECRetry(void)
{
  FINDX last;
  RPREC prec;

  last = SeqLastIdx();
  prec = RulePrec(RevRule());
  /* Choose which index to expand and the associated rule */
  if (RevIdx() != last) {
    /* Since the reversed index was not the rightmost one, there may be 
       indexes on its right hand side having the same precedence. 
       Thus, we try to find the nearest index on the right with precedence 
       equal to that of the reversed index. 
    */ 
    FindNextPrec(prec,SeqNextIdx(RevIdx()),last);
    if ((select_fnidx != FINDX_OUT) && (RulePrec(select_rule) == prec))
      return; /* We have found an index on the right with same precedence */
  }
  /* When we get here, we know, since the scanning is left to right, that 
     the reversed-index was the right-most index in its precedence level.
  */
  if (prec > RPREC_MIN)
    /* We have not treated all precedence levels yet, thus, we must try 
       to find the left-most index whose precedence is lower. */
    FindNextPrec(prec - RPREC_C(1), SeqFirstIdx(), last);
  else 
    /* No more precedence level to be treated */
    select_fnidx = FINDX_OUT;
}

void RPRECInit(void)
/* Initialize rule precedences */
{
  /* CAUTION: both premises of conjunctive and disjunctive rules MUST be 
     given the same precedence. Not doing so can cause the selection routine 
     to conclude that all nodes have been treated although it's not the case.
     Therefore, the proof-search may not examine the whole search-space and 
     give erroneous results about provability.      
  */

  /* Lowest precedence for totally non-invertible rules */
  RuleSetPrec(or_s_imp, RPREC_C(0));
  RuleSetPrec(r_or, RPREC_C(0));
  /* Semi-invertible rule */
  RuleSetPrec(imp_imp, RPREC_C(1));
  /* Invertible rules with two premises */
  RuleSetPrec(l_or, RPREC_C(2)); 
  RuleSetPrec(and_s_imp, RPREC_C(2));
  RuleSetPrec(r_and, RPREC_C(2));
  /* Invertible rules with one premiss */
  RuleSetPrec(imp_s_imp, RPREC_C(3));
  RuleSetPrec(or_imp, RPREC_C(3));
  RuleSetPrec(and_imp, RPREC_C(3));
  RuleSetPrec(r_imp, RPREC_C(3));
  RuleSetPrec(l_and, RPREC_C(3));
  /* Invertible rule cutting whole subtrees */
  RuleSetPrec(atom_imp, RPREC_C(4));
  /* Highest precedence for tree expansion-rule */
  RuleSetPrec(l_imp, RPREC_C(5));
}

BOOL RPRECOkCM(void)
{
  return TRUE; 
  /* With the above precedence list, because invertible rules have higher 
     precedence than non-invertible ones and are therefore  applied before 
     non-invertible rules.
  */
}





