#include "refuter.h"
#include "tracer.h"

/******************
 * IMPLEMENTATION *
 ******************/


/* KRIPKE-TREE BUILDING FUNCTIONS */

void RefuteStart(void)
{
  CheckStart();
  DelKTree(KripkeTree());
  kripke_tree = KTREE_EMPTY;
}

void RefuteExpandFail(void) 
{
  CheckExpandFail();

  /* The forest has been expanded to an irreducible non-axiomatic forest.
     Therefore, we create a one node kripke-tree in which all the atoms 
     in the context of the local sequent (i.e local forest) are forced. 
  */
  CHECK(KTreeIsEmpty(TopKTree()));
  PE_KTREE(Top()) = NewKTree();
}

void RefuteExpandSucc(void) 
{ 
  CheckExpandSucc();

  /* The forest has been expanded to an axiom. 
     There is no kripke-tree to build. */
  CHECK(KTreeIsEmpty(TopKTree()));
  PE_KTREE(Top()) = KTREE_EMPTY;
}

void RefutePremissSucc(void)
{
  CheckPremissSucc();

  /* The rule we have reversed succeeded.
     We have nothing to do at the local level. There can either
     be no kripke-tree yet at the local level (and then all rules
     which have been applied succeeded) or there can be a non-empty
     local kripke-tree (saying that previous non-invertible rules
     failed before our reversed rule succeeded). Anyway, the
     situation is only partially provable for the moment and if it
     happens, later on, that the situation is in fact unprovable
     we will need all the kripke-trees built by the rules that failed
     and that have been stored as branches of the local level kripke-tree. 
     Therefore we must not delete the local kripke-tree when we reverse
     a rule that succeeds.
     On the other hand, what we must do, is to delete the kripke-tree
     returned by the reversed rule since the situation to which that 
     rule leads is provable (because the rule succeeded) and the returned
     kripke-tree is therefore irrelevant. The reason why a rule that 
     succeeds can return a kripke-tree is that we can have expanded
     several non-invertible rules that failed (thus building partial
     local kripke-trees) before finding the rule that leads to a proof. 
  */
  DelKTree(RevKTree());
}

void RefutePremissFail(void)
{
  CheckPremissFail();

  if (RuleIsInver(RevRule(),RevPrem())) {
    /* The premiss is invertible, therefore, the kripke-tree returned by
       the premiss is also a counter-model for the current sequent.
       Consequently, we simply delete the current kripke-tree (if any)
       and replace it by the kripke-tree returned by the premiss.
       CAUTION:
       One could think that deleting the current kripke-tree is useless 
       because, as counter-model construction requires applying invertible 
       rules first, it cannot be the case that an invertible rule fails 
       while a kripke-tree already exists for the current sequent, for it 
       would mean that a non-invertible premiss failed before the reversed
       premiss, which is a contradiction.
       However, applying non-invertible rules first does not imply applying
       non-invertible premises first. For example, ->-> is not invertible
       because one of its premiss is invertible but the other is not. Thus,
       if the non-invertible premiss is applied before the invertible one,
       then, we already have the kripke-tree returned by the non-invertible 
       premiss when the invertible one is reversed.
    */ 
    DelKTree(TopKTree());
    PE_KTREE(Top()) = RevKTree();
  } else {
    /* The premiss is not invertible, therefore, we must gather the
       kripke-tree returned by this premiss with the others kripke-trees
       returned by the previous non-invertible premises that failed.
       Those kripke-trees are gathered as sons of a kripke-node which 
       is the root of the counter-model for the current sequent. 
       If that root-node does not exist yet, it is created before
       the combination takes place.
    */
    if (KTreeIsEmpty(TopKTree()))
      PE_KTREE(Top()) = NewKTree();
    PE_KTREE(Top()) = AddKTree(TopKTree(),RevKTree());
  }
}

void RefuteSuccess(void)
{
  CheckSuccess();

  kripke_tree = TopKTree(); /* Should be KTREE_EMPTY */

  ENSURE(KTreeIsEmpty(KripkeTree()));
}

void RefuteFailure(void)
{
  CheckFailure();

  kripke_tree = TopKTree(); /* Retrieve kripke-tree from the stack */
  kripke_typmet = SearchMethod(); /* Keep track of the construction method */
  kripke_size = KTreeSize(KripkeTree()); /* Compute the kripke-tree size */
  kripke_depth = KTreeDepth(KripkeTree()); /* Compute the kripke-tree depth */

  ENSURE(! KTreeIsEmpty(KripkeTree()));
  ENSURE(KripkeMethod() == SearchMethod());
  ENSURE(KripkeSize() > KSIZE_C(0));
  ENSURE(KTreeIsCM(KripkeTree()));
}
