#ifndef __RULES_H
#define __RULES_H

#include "implem.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* Information needed to reverse some of the premises */

typedef FINDX INFO[4];

/* CAUTION: CONCLU and FSTSON information can share the same index because
   they are not needed together at the same time by any expansion rule. */
#define INFO_TAGGED(i) ((i)[0])
#define INFO_CONCLU(i) ((i)[1])
#define INFO_FSTSON(i) ((i)[1])
#define INFO_LMLEAF(i) ((i)[2])
#define INFO_RMLEAF(i) ((i)[3])

typedef FINDX (*TAPPLY)(FINDX,INFO);
typedef void  (*TREVER)(FINDX,INFO);

/* Data structure for an inference rule */

typedef enum { SINGLE, CONJUNCT, DISJUNCT } RTYPE;

typedef unsigned char RPREC;
#define RPREC_C(c) (c ## U)
#define RPREC_MIN  RPREC_C(0)  /* Lowest precedence */
#define RPREC_MAX  RPREC_C(12) /* Highest precedence */ 
#define RPREC_OUT  RPREC_C(13) /* Invalid precedence */
#define IS_RPREC(p) ((p) <= RPREC_MAX) 
/* Return: TRUE if 'p' is a valid precedence */

typedef struct rule_s rule_t;
struct rule_s
{
  char *name; /* Name of the rule */
  RTYPE type; /* Type of the rule (single,conjunctive,disjunctive) */
  RPREC prec; /* Precedence of the rule */
  struct {
    BOOL inver; /* TRUE if the premiss is invertible */
    TAPPLY apply; /* How to apply the premiss */
    TREVER rever; /* How to reverse the premiss */
  } prem[2];
};

typedef unsigned char RULE;
#define RULE_C(c) (c ## U)
/* Conclusion rules */
#define r_and      RULE_C(0)
#define r_or       RULE_C(1)
#define r_imp      RULE_C(2)
/* Root rules */
#define l_and      RULE_C(3)
#define l_or       RULE_C(4)
#define l_imp      RULE_C(5)
/* Non-starred-leaf rules */
#define and_imp    RULE_C(6)
#define or_imp     RULE_C(7)
#define imp_imp    RULE_C(8)
#define atom_imp   RULE_C(9)
/* Starred-leaf rules */
#define and_s_imp  RULE_C(10)
#define or_s_imp   RULE_C(11)
#define imp_s_imp  RULE_C(12)
#define RULE_OUT   RULE_C(13)
#define RULE_MIN   RULE_C(0)
#define RULE_MAX   RULE_C(12)

#define IS_RULE(r) ((r) <= RULE_MAX)
/* Return: TRUE of 'r' is a valid rule */

typedef unsigned char PREM;
#define PREM_FST 0U
#define PREM_SND 1U
#define PREM_OUT 2U
#define PREM_MIN PREM_FST
#define PREM_MAX PREM_SND

#define IS_PREM(p) ((p) <= PREM_SND)
/* Return: TRUE if 'p' is a valid premiss */

implem rule_t trule[];  /* Array of rules */

#define RU(r) trule[(r)]
#define RU_PR(r,p) (RU(r).prem[(p)])


/*************
 * INTERFACE *
 *************/

/* CONSTRUCTORS - DESTRUCTORS */

public void MakeRules(void);
/* Initialize all logical rules */

/* ACCESSORS */

#define RuleName(r) GETSAFE(IS_RULE(r),RU(r).name)
/* Return: the name of rule 'r' */

#define RulePrec(r) GETSAFE(IS_RULE(r),RU(r).prec)
/* Return: the precedence of rule 'r' */

#define RuleType(r) GETSAFE(IS_RULE(r),RU(r).type)
/* Return: the type of rule 'r' */

#define RuleIsSing(r) (RuleType(r) == SINGLE)
/* Return: TRUE if 'r' is a single premiss rule */

#define RuleIsConj(r) (RuleType(r) == CONJUNCT)  
/* Return: TRUE if 'r' is a conjunction of two premises */

#define RuleIsDisj(r) (RuleType(r) == DISJUNCT)
/* Return: TRUE if 'r' is a disjunction of two premises */
  
#define RuleArity(r)  (RuleIsSing(r) ? 1:2)
/* Return: the arity of rule 'r' between 1 and 2 */

#define RuleIsPrem(r,p) ((RuleIsSing(r) && ((p) == PREM_FST)) || IS_PREM(p))
/* Return: TRUE if 'r' is a valid rule and 'p' a valid premiss of 'r' */ 

#define RuleIsInver(r,p) GETSAFE(RuleIsPrem((r),(p)),RU_PR((r),(p)).inver)
/* Return: TRUE if premiss of index 'p' of rule 'r' is invertible */

public RULE FindRule(FINDX index);
/* Return: the rule which can be applied to 'index' or 
           RULE_OUT if no rule can be applied.
   Require:
     Except for the conclusion, 'index' is supposed to be effectively 
     chained in the list of forest leaves.
*/

#define PremSwitch(p) GUARD(IS_PREM(p),(PREM_FST + PREM_SND) - (p))
/* Return: PREM_FST if 'p' is equal to PREM2 and vice versa */

/* MODIFIERS */

#define RuleApply(r,p,idx,inf) GUARD(RuleIsPrem((r),(p)),\
                                     RU_PR((r),(p)).apply((idx),(inf)))


#define RuleReverse(r,p,idx,inf) GUARD(RuleIsPrem((r),(p)),\
				       RU_PR((r),(p)).rever((idx),(inf)))

#define RuleSetPrec(r,n) SETSAFE(IS_RULE(r) && IS_RPREC(n),RU(r).prec,(n))

#endif


