#include "strip.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* SEARCH INPUTS */
LFORM forest_lform; /* The list of context formulae of the current forest */
PFORM forest_concl; /* The conclusion formula of the current forest */
bmask_t  use_bmask; /* How to solve the proof-search: check,prove,refute */
typtrc_t use_typtrc; /* How to trace the proof-search */
typmet_t use_typmet; /* The proof-search method  to use */
typimp_t use_typimp; /* The forest implementation to use */
unsigned int use_nbiter; /* How many check iterations to perform */

/* SEARCH STATE */

FOREST forest; /* PRIVATE: the proof-search forest */
IMPLEM forimp; /* PRIVATE: the proof-search forest-implementation */
ubint_t forimp_base; /* Cost of the forest operations in the ideal case */
ubint_t forimp_ovhd; /* Cost of the forest operations overhead */ 
PSTACK pstack; /* PRIVATE: the proof-search stack */
FINDX focus_fnidx; /* A valid index in the current forest to start with */
METHOD method; /* PRIVATE: the proof-search method */
ubint_t method_base; /* Cost of the method in the ideal case */
ubint_t method_ovhd; /* Cost of the method overhead */
FINDX select_fnidx; /* Selected index to apply (returned by the method) */ 
RULE  select_rule; /* Selected rule associated with "select_fnidx" */
PREM  select_prem; /* Index of the selected premiss */
ENGINE engine; /* PRIVATE: the proof-search automaton */

/* SEARCH PARAMETERS */
typimp_t search_typimp; /* The forest implementation used for the result */
typmet_t search_typmet; /* The method used for the result */  

/* SEARCH RESULTS */
unsigned int search_iter; /* The number of check loops used for the result */
milsec_t search_time; /* How long it took to decide the current forest */
typres_t search_typres; /* The result of the proof-search */

/* CHECKER RESULTS */
ubint_t  check_depth; /* Max depth reached during proof-search */
ubint_t  check_nbapp; /* Number of rules applied during proof-search */
ubint_t  check_pmem; /* Number of proof-nodes for proof-construction */
ubint_t  check_kmem; /* Number of kripke-nodes for cm-construction */
ubint_t  check_mbase; /* Basic cost of the check method */
ubint_t  check_movhd; /* Overhead cost of the check method */
ubint_t  check_fbase; /* Basic cost of the check forest-implem */
ubint_t  check_fovhd; /* Overhead cost of the check forest-implem */

/* PROOF-BUILDER RESULTS */
PTREE    proof_tree; /* Pointer to the root of the proof-tree */
typmet_t proof_typmet; /* The method used to build the proof */
ubint_t  proof_size;   /* Size of the proof */  
ubint_t  proof_depth;  /* Depth of the proof */

/* COUNTER-MODEL BUILDER RESULTS */
KTREE    kripke_tree; /* Pointer to the root of the current kripke-tree */
typmet_t kripke_typmet; /* The method used to build the kripke-tree */
ubint_t  kripke_size; /* Number of nodes in the kripke-tree */
ubint_t  kripke_depth; /* Depth of the kripke-tree */

/* STRIP IMPLEMS */
implem_t timp[] = { DEF_IMPLEM(ISC), DEF_IMPLEM(LRM) };
char *typimp_t_str[] = { "isc", "lrm" };

/* STRIP METHODS */
method_t tmet[] = { DEF_METHOD(RPREC), DEF_METHOD(FLEAF) } ;
char *typmet_t_str[] = { "rprec"," fleaf" };

/* STRIP ENGINES */
engine_t checker = DEF_ENGINE(Check);
engine_t tracer  = DEF_ENGINE(Trace);
engine_t prover  = DEF_ENGINE(Prove);
engine_t refuter = DEF_ENGINE(Refute);

/* MISCELLANEOUS */
char *typres_t_str[] = { "unprovable","provable" };
char *typtrc_t_str[] = { "normal","prompt" };
char *bmask_t_str[] = { "check", "prove", "refute", "prove | refute" };

/***********************************************************************/

/* FUNCTIONS */

/* HELPERS */

void ResetSettings(void)
{
  SetImplem(TYPIMP_ISC);
  SetMethod(TYPMET_RPREC);
  SetTraceMode(TYPTRC_NORMAL);
  SetBuildMask(BMASK_NOTHING);
  SetIterator(1U);
}

void ResetResults(void)
{
  search_iter = 0U;
  search_time = MILSEC_C(0);
  search_typres = TYPRES_OUT;
  check_depth = UBINT_C(0);
  check_nbapp = UBINT_C(0);
  check_pmem = UBINT_C(0);
  check_kmem = UBINT_C(0);
  method_base = UBINT_C(0);
  method_ovhd = UBINT_C(0);
  forimp_base = UBINT_C(0);
  forimp_ovhd = UBINT_C(0);
}

void KillProof(void)
{
  FreePSpace();
  proof_tree = PTREE_EMPTY;
  proof_size = UBINT_C(0);
  proof_depth = UBINT_C(0);
}

void KillCM(void)
{
  KillKSpace();
  kripke_tree = KTREE_EMPTY;
  kripke_size = KSIZE_C(0);
  kripke_depth = KSIZE_C(0);
}

/* MODIFIERS */

void SetIterator(unsigned int n)
{
  if (n < 1U)
    use_nbiter = 1U;
  else if (n > MAX_ITER)
    use_nbiter = MAX_ITER;
  else
    use_nbiter = n;
}

void SetMethod(typmet_t met)
{
  REQUIRE(IS_TYPMET(met));
  use_typmet = met;
  ENSURE(Method() == met);
}

void SetImplem(typimp_t imp)
{
  REQUIRE(IS_TYPIMP(imp));
  use_typimp = imp;
  ENSURE(Implem() == imp);
}

void Define(LFORM l,PFORM c)
{
  static BOOL once;

  if (! once)
    once = TRUE;
  else {
    /* We have already defined something. We must do some clean-up 
       before we define something else. */
    KillPs();
    KillForest(); 
    KillProof();
    KillCM();
  }

  MakeForest(l,c); /* New forest for the sequent */
  MakePs(); /* New proof-search stack */
  ResetResults();
}

BOOL Check(unsigned int n)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST); /* Error: no forest has been defined */

  search_typmet = Method();
  search_typimp = Implem();
  engine = &checker;
  Search(n);
  RetOk();
}

BOOL Trace(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST); /* Error: no forest has been defined */

  search_typmet = Method();
  search_typimp = Implem();
  engine = &tracer;
  Search(1); 
  RetOk();
}

BOOL BuildProof(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST); /* Error: no forest has been defined */

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  if (SearchResult() != SUCCESS)
    RetErr(ERR_NOPROOF); /* Error: the sequent is not provable */

  if (! PTreeIsEmpty(ProofTree())) {
    /* We already have a proof. Thus, we must check whether this proof 
       has been obtained with the same method as the current result.
       If it is not the case, the proof must be rebuilt according to 
       the current result method since the current proof-size estimation 
       is only relevant for that method. 
    */    
    if (ProofMethod() == SearchMethod())
      RetOk();
    else
      KillProof(); /* Free the memory used by the current proof */ 
  }
  /* We arrive here if we need to (re)build a proof. For that, we must 
     first allocate enough memory to hold the proof according to the 
     proof-size information computed during the proof-search. 
  */
  CHECK(PTreeIsEmpty(ProofTree()));
  CHECK(CheckPMem() > UBINT_C(0));

  if (! AllocPSpace(CheckPMem()))
    RetErr(ERR_NOMEMORY); /* Error: not enough memory */

  /* Build the proof */
  engine = &prover;
  Search(1);

  /* We should now have a suitable proof */
  ENSURE((! PTreeIsEmpty(ProofTree())) && (ProofMethod() == SearchMethod()));
  RetOk();
}

BOOL Prove(void)
{
  if (! Check(1)) return FALSE;
  if (SearchResult() == SUCCESS)
    return BuildProof();
  RetOk();
}

BOOL BuildKripke(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST); /* Error: no forest has been defined */

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  if (SearchResult() != FAILURE)
    RetErr(ERR_NOMODEL); /* Error: the sequent is provable */

  if (! KTreeIsEmpty(KripkeTree()))
    /* We already have a counter-model. Thus, we must check whether this 
       counter-model has been obtained with the same method as the result.
       If it is not the case, the counter-model must be rebuilt (if possible) 
       according to the result method since the kripke space-size estimation 
       is only relevant for that method. 
    */    
    if (KripkeMethod() == SearchMethod())
      RetOk(); /* Keep the current counter-model */
  
  if (MethodOkCM())
    /* The method allow cm-construction, kill the current counter-model. */
    KillCM();
  else if (! KTreeIsEmpty(KripkeTree()))
    /* We already have a counter-model but not the one corresponding to
       the result method which, anyway, cannot be used to obtain one. 
       So, let's keep the one we have. */
    RetOk();
  else /* We don't have any counter-model available. */
    RetErr(ERR_BADMETCM); /* Error: bad method for cm-construction */

  /* We arrive here if we can (re)build a counter-model. For that, we must 
     first allocate enough memory to hold the counter-model, according to 
     the counter-model size information computed during the proof-search. 
  */
  CHECK(KTreeIsEmpty(KripkeTree()));
  CHECK(CheckKMem() > UBINT_C(0));

  if (! MakeKSpace(CheckKMem()))
    RetErr(ERR_NOMEMORY); /* Error: not enough memory */

  /* Build the counter-model */
  engine = &refuter;
  Search(1);

  /* We should now have a suitable counter-model if the method allowed it */
  ENSURE(KTreeIsEmpty(KripkeTree()) || (KripkeMethod() == SearchMethod()));
  RetOk();
}

BOOL Refute(void)
{
  if (! Check(1)) return FALSE;
  if (SearchResult() == FAILURE)
    return BuildKripke();
  RetOk();
}

BOOL Build(void)
{
  switch(SearchResult()) {
  case SUCCESS: 
    if (BuildMask() & BMASK_PROOF) 
      return BuildProof();
    else
      return TRUE;
  case FAILURE: 
    if (BuildMask() & BMASK_KRIPKE)
      return BuildKripke();
    else
      return TRUE;
  }
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST);
  else
    RetErr(ERR_NORESULT);
}

BOOL Solve(void)
{
  if (! Check(1)) return FALSE; 
  return Build();
}

/* CONSTRUCTORS - DESTRUCTORS */

void MakeSTRIP(void)
{
  MakeRules();
  ResetSettings();
  ResetResults();
  KillProof();
  KillCM();
}

void KillSTRIP(void)
{
  KillPs();
  KillForest();
  KillProof();
  KillCM();
}
