#ifndef __STRIP_H
#define __STRIP_H

#include "checker.h"
#include "tracer.h"
#include "prover.h"
#include "refuter.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* STRIP ERRORS */
#define ERR_NOFOREST  ERR_C(1)  /* No defined sequent */
#define ERR_NORESULT  ERR_C(2)  /* Unsolved sequent */
#define ERR_NOHELP    ERR_C(3)  /* No help file available */
#define ERR_NOPROOF   ERR_C(4)  /* No proof available */
#define ERR_NOMODEL   ERR_C(5)  /* No model available */
#define ERR_NOMEMORY  ERR_C(6)  /* Not enough memory */
#define ERR_BADMETCM  ERR_C(7)  /* Bad method for counter-models */
#define ERR_UNPROVED  ERR_C(8)  /* Unproved sequent */
#define ERR_UNREFUTED ERR_C(9)  /* Unrefuted sequent */
#define ERR_IO        ERR_C(10) /* IO error */

#define MAX_ITER UINT_MAX

/* MISCELLANEOUS */

typedef enum { BMASK_NOTHING,BMASK_PROOF,BMASK_KRIPKE,BMASK_BOTH } bmask_t;

#define IS_BMASK(bm) ((bm) <= BMASK_BOTH)

/*************
 * INTERFACE *
 *************/

/* CONSTRUCTORS AND DESTRUCTORS */

public void MakeSTRIP(void); 
/* Construct and initialize the STRIP theorem prover */

public void KillSTRIP(void); 
/* Destroy the STRIP theorem prover */

/* ACCESSORS */

/* Settings */

implem unsigned int use_nbiter;

#define Iterator() VALUE(use_nbiter)
/* Return: the number of proof-search iterations to perform */

implem bmask_t use_bmask;

#define BuildMask() VALUE(use_bmask)
/* Return: the current value of the build mask */

implem char *bmask_t_str[];

#define BuildMaskStr() GETSAFE(IS_BMASK(BuildMask()),\
			       bmask_t_str[BuildMask()])
/* Return: the name of the current solve mode in clear */

implem typmet_t use_typmet;

#define Method() VALUE(use_typmet)
/* Return: the current proof-search method */

implem typimp_t use_typimp;

#define Implem() VALUE(use_typimp)
/* Return: the current forest implem */

size_t GetStripDir(char *dest,size_t max);

/* MODIFIERS */

public void ResetSettings(void);
/* Sets the default settings */

public void SetIterator(unsigned int n);
/* Define 'n' as the current number of check iterations
   Remarks
     If 'n' < 1, the number of iterations is set to 1
     If 'n' > MAX_ITER, the number of iterations is set to MAX_ITER */

public void SetMethod(typmet_t met);
/* Set 'met' as the current proof-search method */

public void SetImplem(typimp_t imp);
/* Set 'imp' as the current forest implementation */

#define SetBuildMask(mask) SETSAFE(IS_BMASK(mask),use_bmask,(mask))
/* Define 'mode' as the current solve mode value */

public void Define(LFORM antecedent,PFORM conclusion);
/* Define the sequent 'antecedent' => 'conclusion' as the current sequent */

public BOOL Check(unsigned int n);
/* Decide the current sequent and repeat the proof-search 'n' times
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if no sequent has been defined */

public BOOL Trace(void);
/* Trace the proof-search.
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if no sequent has been defined. */

public BOOL BuildProof(void);
/* Build a proof out of the proof-search.
   Return
     TRUE if the function succeeds. 
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_NOPROOF: if the sequent has no proof 
       ERR_NOMEMORY: if there is not enough memory to build the proof */

public BOOL Prove(void);
/* Check the current sequent and build a proof if provable
   Return
     TRUE if the function succeeds. 
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NOMEMORY: if there is not enough memory to build the proof */

public BOOL BuildKripke(void);
/* Build a counter-model out of the proof-search.
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_NOMODEL: if the sequent has no counter-model
       ERR_NOMEMORY: if there is enough memory to build the counter-model
       ERR_BADMETCM: if the method does allow counter-model construction */

public BOOL Refute(void);
/* Check the current sequent and build a counter-model if unprovable
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NOMEMORY: if there is enough memory to build the counter-model
       ERR_BADMETCM: if the method does allow counter-model construction */

public BOOL Build(void);
/* Build a proof or counter-model out of the proof-search,
   according on the build mask.
   Return
     TRUE if the function succeeds. 
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_NOPROOF: if the sequent has no proof 
       ERR_NOMODEL: if the sequent has no counter-model
       ERR_NOMEMORY: if there is not enough memory to build the proof
       ERR_BADMETCM: if the method does allow counter-model construction */


public BOOL Solve(void);
/* Decide the current sequent and build a proof or a counter-model out of 
   the proof-search, depending on the search result and the build mask.
   Return
     TRUE if the function succeeds. 
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_NOPROOF: if the sequent has no proof 
       ERR_NOMODEL: if the sequent has no counter-model
       ERR_NOMEMORY: if there is not enough memory to build the proof
       ERR_BADMETCM: if the method does allow counter-model construction */


/* DISPLAY */

public BOOL Help(void);
/* Return
     TRUE if the function succeeds
     The function fails with error
       ERR_NOHELP: if no help file is available */

public void PrintErr(void);
/* Print a message corresponding to the current error on 'stdout' */

public BOOL PrintSet(void);
/* Print the current settings on 'stdout' */

public BOOL VerboseSet(void);
/* Print the current settings on 'stdout' in details */
 
public BOOL PrintResult(void);
/* Return 
     TRUE if the function succeeds 
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet */

public BOOL PrintKripke(void);
/* Print the current counter-model on 'stdout'
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_UNREFUTED: if the sequent has not been refuted yet
       ERR_NOMODEL: if the sequent has no counter-model
*/

public BOOL PrintProof(void);
/* Print the current proof on 'stdout' 
   Return
     TRUE if the function succeeds.
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet
       ERR_UNPROVED: if the sequent has not been proved yet
       ERR_NOPROOF: if the sequent has no proof
*/

public BOOL PrintStat(void);
/* Return 
     TRUE if the function succeeds
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet */

public BOOL VerboseStat(void);
/* Return 
     TRUE if the function succeeds
     The function fails with error
       ERR_NOFOREST: if there is no defined sequent
       ERR_NORESULT: if the sequent has not been decided yet */

public BOOL PrintMP(void);
/* Return: */

public BOOL PrintTEX(void);
/* Return: */

public BOOL PrintXML(void);
/* Return: */

public BOOL PrintGDL(void);
/* Return: */

#endif
